import urllib.parse
from typing import List

from spinta import exceptions
from spinta.components import UrlParseNode
from spinta.spyna import unparse

RULES = {
    'path': {
        'minargs': 0,
    },
    'ns': {
        'maxargs': 0,
    },
    # Recursivelly return all objects under that namespace.
    'all': {
        'maxargs': 0,
    },
    'changes': {
        'minargs': 0,
        'maxargs': 1,
    },
    'format': {
        'maxargs': 1,
    },
    # In batch requests, return summary of what was done.
    'summary': {
        'maxargs': 2,
    },
    # In batch requests, continue execution even if some actions fail.
    'fault-tolerant': {
        'maxargs': 0,
    },
    # Wipe all data of a model or all models in a namespace.
    'wipe': {
        'maxargs': 0,
    },
    # Check manifest.
    'check': {
        'maxargs': 0,
    }
}


def apply_query_rules(rules, params):
    for param in params:
        name = param['name']
        args = param['args']

        if name not in rules:
            raise exceptions.UnknownRequestParameter(name=name)

        rule = rules[name]
        maxargs = rule.get('maxargs')
        minargs = rule.get('minargs', 0 if maxargs == 0 else 1)
        if len(args) < minargs:
            # FIXME: This should be an UserError.
            raise Exception(
                f"At least {minargs} argument is required for {name!r} URL "
                "parameter."
            )

        if maxargs is not None and len(args) > maxargs:
            # FIXME: This should be an UserError.
            raise Exception(
                f"URL parameter {name!r} can only have {maxargs} arguments."
            )
    return params


def parse_url_path(path) -> List[UrlParseNode]:
    query = []
    name = None if path.startswith(':') else 'path'
    args = []
    parts = map(urllib.parse.unquote, path.split('/')) if path else []
    for part in parts:
        if part.startswith(':'):
            if name is not None:
                query.append({
                    'name': name,
                    'args': args,
                })
            name = part[1:]
            args = []
        else:
            args.append(part)
    query.append({
        'name': name,
        'args': args,
    })
    apply_query_rules(RULES, query)
    return query


def build_url_path(query: List[UrlParseNode]):
    parts = []
    other = []
    for param in query:
        name = param['name']
        args = param['args']
        if name == 'path':
            parts.extend(args)
        elif name in ('format', 'ns', 'changes'):
            parts.extend([f':{name}'] + args)
        else:
            other.append(param)
    path = '/'.join(map(str, parts))
    if other:
        other = unparse({'type': 'expression', 'name': 'and', 'args': other})
        return f'{path}?{other}'
    else:
        return path



import datetime
import uuid
from typing import Dict

import contextlib

from spinta.typing import ObjectData
from spinta.backends import Backend
from spinta.backends.components import BackendFeatures


class NoBackend(Backend):
    name = 'no backend'
    metadata = {
        'name': 'no backend',
    }

    def __init__(self):
        self.data = {}

    @contextlib.contextmanager
    def begin(self):
        yield self

    def bootstrapped(self):
        return True



from typing import AsyncIterator

from spinta import commands
from spinta.components import Context, Property, DataItem
from spinta.backends.fs.components import FileSystem


@commands.create_changelog_entry.register(Context, Property, FileSystem)
def create_changelog_entry(
    context: Context,
    model: Property,
    backend: FileSystem,
    *,
    dstream: AsyncIterator[DataItem],
) -> AsyncIterator[DataItem]:
    # FileSystem properties don't have change log, change log entry will be
    # created on property model backend.
    return dstream



from typing import List

from spinta import commands
from spinta.components import Context
from spinta.core.access import link_access_param
from spinta.datasets.components import Dataset, Resource, Entity, Attribute
from spinta.exceptions import MissingReference
from spinta.types.datatype import Partial, Ref
from spinta.components import Property


@commands.link.register(Context, Dataset)
def link(context: Context, dataset: Dataset):
    link_access_param(dataset, (dataset.manifest,))
    for resource in dataset.resources.values():
        commands.link(context, resource)


@commands.link.register(Context, Resource)
def link(context: Context, resource: Resource):
    link_access_param(resource, (resource.dataset,))


@commands.link.register(Context, Entity)
def link(context: Context, entity: Entity):
    datasets = entity.model.manifest.datasets
    if entity.dataset:
        if entity.dataset not in datasets:
            raise MissingReference(
                entity,
                param='dataset',
                ref=entity.dataset,
            )
        # XXX: https://gitlab.com/atviriduomenys/spinta/-/issues/44
        dataset: str = entity.dataset
        entity.dataset = datasets[dataset]

        resources = entity.dataset.resources
        if entity.resource:
            if entity.resource not in resources:
                raise MissingReference(
                    entity,
                    param='resource',
                    ref=entity.resource,
                )
            # XXX: https://gitlab.com/atviriduomenys/spinta/-/issues/44
            resource: str = entity.resource
            entity.resource = resources[resource]

            assert entity.model.name not in entity.resource.models
            entity.resource.models[entity.model.name] = entity.model
        else:
            entity.resource = None
    else:
        entity.dataset = None


@commands.link.register(Context, Attribute)
def link(context: Context, attribute: Attribute):
    pass



import base64

from spinta import commands
from spinta.utils.schema import NA
from spinta.components import Context
from spinta.types.datatype import File
from spinta.types.helpers import check_no_extra_keys
from spinta.formats.components import Format
from spinta.formats.json.components import Json
from spinta.backends.components import Backend


@commands.load.register(Context, File, object)
def load(context: Context, dtype: File, value: object) -> object:
    if value is NA:
        return value

    assert isinstance(value, dict)

    # check that given obj does not have more keys, than dtype's schema
    check_no_extra_keys(dtype, dtype.schema, value)

    if '_content' in value and isinstance(value['_content'], str):
        value['_content'] = base64.b64decode(value['_content'])

    return commands.decode(context, Json(), dtype.backend, dtype, value)


@commands.decode.register(Context, Format, Backend, File, dict)
def decode(context: Context, source: Format, target: Backend, dtype: File, value: dict):
    return value



import os
import pathlib
import tempfile
from difflib import Differ
from typing import Any, Iterator

import pprintpp
import pytest
import sqlalchemy as sa
import sqlalchemy_utils as su
from responses import RequestsMock

from spinta.core.config import RawConfig
from spinta.core.config import read_config
from spinta.manifests.components import Manifest
from spinta.testing.cli import SpintaCliRunner
from spinta.testing.client import TestClient
from spinta.testing.client import create_test_client
from spinta.testing.context import ContextForTests
from spinta.testing.context import create_test_context
from spinta.testing.datasets import Sqlite
from spinta.testing.manifest import compare_manifest


def _remove_push_state(rc: RawConfig) -> None:
    data_dir = rc.get('data_path')
    for file in (data_dir / 'push').glob('*.db'):
        file.unlink()


@pytest.fixture(scope='session')
def rc():
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = pathlib.Path(tmpdir)
        data_dir = tmpdir / '.local/share'
        rc = read_config()
        rc.add('pytest', {
            'env': 'test',
            'data_path': data_dir,
            'keymaps.default': {
                'type': 'sqlalchemy',
                'dsn': 'sqlite:///{data_dir}/keymap.db',
            },
        })
        rc.lock()
        yield rc
        _remove_push_state(rc)


@pytest.fixture()
def sqlite():
    with tempfile.TemporaryDirectory() as tmpdir:
        yield Sqlite('sqlite:///' + os.path.join(tmpdir, 'db.sqlite'))


def _prepare_postgresql(dsn: str) -> None:
    engine = sa.create_engine(dsn)
    with engine.connect() as conn:
        conn.execute(sa.text("CREATE EXTENSION IF NOT EXISTS postgis"))
        conn.execute(sa.text("CREATE EXTENSION IF NOT EXISTS postgis_topology"))
        conn.execute(sa.text("CREATE EXTENSION IF NOT EXISTS fuzzystrmatch"))
        conn.execute(sa.text("CREATE EXTENSION IF NOT EXISTS postgis_tiger_geocoder"))


@pytest.fixture(scope='session')
def postgresql(rc) -> str:
    dsn: str = rc.get('backends', 'default', 'dsn', required=True)
    if su.database_exists(dsn):
        _prepare_postgresql(dsn)
        yield dsn
    else:
        su.create_database(dsn)
        _prepare_postgresql(dsn)
        yield dsn
        su.drop_database(dsn)


@pytest.fixture(scope='session')
def mongo(rc):
    yield
    dsn = rc.get('backends', 'mongo', 'dsn', required=False)
    db = rc.get('backends', 'mongo', 'db', required=False)
    if dsn and db:
        import pymongo
        client = pymongo.MongoClient(dsn)
        client.drop_database(db)


@pytest.fixture(scope='session')
def backends(postgresql, mongo):
    yield {
        'postgresql': postgresql,
        'mongo': mongo,
    }


@pytest.fixture(scope='session')
def _context(rc: RawConfig, postgresql, mongo):
    context: ContextForTests = create_test_context(rc)
    context.load()
    yield context


@pytest.fixture
def context(_context, mocker, tmp_path, request):
    with _context.fork('test') as context:
        store = context.get('store')
        if 'fs' in store.backends:
            # XXX: There must be a better way to provide tmpdir to fs backend.
            mocker.patch.object(store.backends['fs'], 'path', tmp_path)

        # In-memory accesslog used with spinta.accesslog.python.
        context.set('accesslog.stream', [])

        yield context

        # At this point, transaction must be closed, if it is not, then something is
        # wrong. Find out why transaction was not property closed.
        assert context.has('transaction') is False

        # If context was not loaded, then it means, that database was not touched.
        # All database operations require fully loaded context.
        if context.loaded:
            # XXX: Maybe instead of deleting everything, we could rollback
            #      transactions, once this kind of functionality will be
            #      available? This should be more efficient.
            context.wipe_all()


@pytest.fixture
def responses() -> Iterator[RequestsMock]:
    with RequestsMock() as mock:
        yield mock


@pytest.fixture
def app(context) -> TestClient:
    context.attach('client', create_test_client, context)
    return context.get('client')


@pytest.fixture
def cli(rc: RawConfig):
    yield SpintaCliRunner(mix_stderr=False)
    _remove_push_state(rc)


def pytest_addoption(parser):
    # TODO: Switch back to backend pytest param.
    #       We want to tests multiple backends and one backend can use
    #       multiple models.
    parser.addoption(
        "--model",
        action="append",
        default=[],
        help="run tests only for particular model ['postgres', 'mongo', 'postgres/datasets']",
    )


def pytest_configure(config):
    # https://docs.pytest.org/en/latest/mark.html#registering-marks
    config.addinivalue_line(
        "markers", "models(*models): mark test to run multiple times with each model specified"
    )


def pytest_generate_tests(metafunc):
    # Get model markers from test, if markers are set - leave test as is
    models = metafunc.definition.get_closest_marker('models')
    if not models:
        return

    # If there are markers, get them, together with model CLI options
    models = set(models.args)
    model_cli_options = set(metafunc.config.getoption('model'))

    # If model CLI options are not empty
    # then get common markers from test and CLI options
    if model_cli_options:
        models = models.intersection(model_cli_options)

    # Parametrize our test with calculated models.
    # If we pass to CLI model option, which does not have a test marker,
    # then pytest will skip the test all together.
    metafunc.parametrize('model', models)


def _diff_line(line: str) -> str:
    if line.startswith('- '):
        return '< ' + line[2:]
    if line.startswith('+ '):
        return '> ' + line[2:]
    return line


@pytest.hookimpl(tryfirst=True)
def pytest_assertrepr_compare(op: str, left: Any, right: Any):
    if op == '==' and isinstance(left, Manifest) and isinstance(right, str):
        left, right = compare_manifest(left, right)
        return ['not equal'] + [
            _diff_line(line)
            for line in Differ().compare(
                left.splitlines(),
                right.splitlines(),
            )
        ]
    types = (dict, list)
    if op == '==' and isinstance(left, types) and isinstance(right, types):
        left = pprintpp.pformat(left, indent=2, width=40).splitlines()
        right = pprintpp.pformat(right, indent=2, width=40).splitlines()
        return ['not equal'] + [
            _diff_line(line)
            for line in Differ().compare(left, right)
        ]



import pathlib

from spinta import commands
from spinta.components import Context
from spinta.manifests.yaml.components import InlineManifest
from spinta.manifests.yaml.components import YamlManifest
from spinta.manifests.yaml.helpers import yaml_config_params


@commands.configure.register(Context, YamlManifest)
def configure(context: Context, manifest: YamlManifest):
    rc = context.get('rc')
    manifest.path = rc.get(
        'manifests', manifest.name, 'path',
        cast=pathlib.Path,
        required=True,
    )
    yaml_config_params(context, manifest)


@commands.configure.register(Context, InlineManifest)
def configure(context: Context, manifest: InlineManifest):
    rc = context.get('rc')
    manifest.path = None
    manifest.manifest = rc.get(
        'manifests', manifest.name, 'manifest',
        required=True,
    )



from typing import Optional

from spinta.core.ufuncs import Expr
from spinta.manifests.components import Manifest
import sqlalchemy as sa


class SqlManifest(Manifest):
    type = 'sql'
    path: Optional[str] = None
    prepare: Optional[Expr] = None

    @staticmethod
    def detect_from_path(path: str) -> bool:
        try:
            url = sa.engine.make_url(path)
            url.get_dialect()
            return True
        except:
            return False



import configparser
import dataclasses
import pathlib
import urllib.parse
from typing import List
from typing import Optional

import requests

from spinta.exceptions import RemoteClientCredentialsNotFound
from spinta.exceptions import RemoteClientCredentialsNotGiven
from spinta.exceptions import RemoteClientScopesNotGiven


@dataclasses.dataclass
class RemoteClientCredentials:
    section: str        # client section in credentials.cfg
    client: str         # client username
    secret: str         # client secret
    server: str         # server URL
    remote: str         # remote name given in credentials.cfg section
    scopes: List[str]   # list of client scopes


def _parse_client_handle(
    handle: str,  # can be server URL or section name in credentials.cfg
) -> RemoteClientCredentials:
    # XXX: Probably this should be dropped, because push should be only
    #      possible to another Spinta or compatible server.
    if handle.startswith('spinta+'):
        handle = handle[len('spinta+'):]

    if handle.startswith(('https://', 'http://')):
        url = handle
    else:
        url = 'https://' + handle

    purl = urllib.parse.urlparse(url)
    section = purl.hostname
    server = f'{purl.scheme}://{purl.hostname}'
    remote = purl.hostname.replace('.', '_').replace('-', '_')
    if purl.port:
        section += f':{purl.port}'
        server += f':{purl.port}'
        if purl.port != 80 and purl.port != 443:
            remote += f'_{purl.port}'
    if purl.username:
        section = f'{purl.username}@{purl.hostname}'
    return RemoteClientCredentials(
        section=section,
        client=purl.username,
        secret=purl.password,
        server=server,
        remote=remote,
        scopes=[],
    )


def get_client_credentials(
    credsfile: Optional[pathlib.Path],  # path to credentials.cfg
    name: str,                  # section name from credentials.cfg or a server
                                # URL
    *,
    check: bool = True,         # check if credentials.cfg exists, section
                                # exists if client username, password and
                                # scopes are given
) -> RemoteClientCredentials:
    creds = _parse_client_handle(name)

    if credsfile:
        if credsfile.exists():
            config = configparser.ConfigParser()
            config.read(credsfile)

            if config.has_section(creds.section):
                creds.server = config.get(creds.section, 'server', fallback=creds.server)
                creds.client = creds.client or config.get(creds.section, 'client', fallback=None)
                creds.secret = creds.secret or config.get(creds.section, 'secret', fallback=None)
                creds.scopes = config.get(creds.section, 'scopes', fallback=[])

            elif check:
                raise RemoteClientCredentialsNotFound(
                    name=name,
                    section=creds.section,
                    credentials=credsfile,
                )

        elif check:
            raise RemoteClientCredentialsNotFound(
                name=name,
                section=creds.section,
                credentials=credsfile,
            )

    if check:
        if not creds.client or not creds.secret:
            raise RemoteClientCredentialsNotGiven(
                name=name,
                section=creds.section,
                credentials=credsfile,
            )

        if not creds.scopes:
            raise RemoteClientScopesNotGiven(
                name=name,
                section=creds.section,
                credentials=credsfile,
            )

    return creds


def add_client_credentials(
    credsfile: pathlib.Path,
    server: str,
    *,
    client: str = None,
    secret: str = None,
    scopes: List[str] = None,
    section: str = None,
) -> None:
    creds = get_client_credentials(credsfile, server, check=False)

    scopes = scopes or creds.scopes
    scopes = '\n' + '\n'.join(f'  {s}' for s in scopes)

    section = section or creds.section

    config = configparser.ConfigParser()
    if credsfile.exists():
        config.read(credsfile)
    config[section] = {
        'server': creds.server,
        'client': client or creds.client or '',
        'secret': secret or creds.secret or '',
        'scopes': scopes,

    }

    with credsfile.open('w') as f:
        config.write(f)


def get_access_token(creds: RemoteClientCredentials) -> str:
    auth = (creds.client, creds.secret)
    resp = requests.post(f'{creds.server}/auth/token', auth=auth, data={
        'grant_type': 'client_credentials',
        'scope': creds.scopes,
    })
    resp.raise_for_status()
    return resp.json()['access_token']



import logging

from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.sql.components import SqlManifest
from spinta.manifests.sql.helpers import read_schema

log = logging.getLogger(__name__)


@commands.load.register(Context, SqlManifest)
def load(
    context: Context,
    manifest: SqlManifest,
    *,
    into: Manifest = None,
    freezed: bool = True,
    rename_duplicates: bool = False,
    load_internal: bool = True,
):
    assert freezed, (
        "SqlManifest does not have unfreezed version of manifest."
    )

    if load_internal:
        target = into or manifest
        if '_schema' not in target.models:
            store = context.get('store')
            commands.load(context, store.internal, into=target)

    schemas = read_schema(context, manifest.path, manifest.prepare)

    if into:
        log.info(
            'Loading freezed manifest %r into %r from %s.',
            manifest.name,
            into.name,
            manifest.path,
        )
        load_manifest_nodes(context, into, schemas, source=manifest)
    else:
        log.info(
            'Loading freezed manifest %r from %s.',
            manifest.name,
            manifest.path,
        )
        load_manifest_nodes(context, manifest, schemas)

    for source in manifest.sync:
        commands.load(
            context, source,
            into=into or manifest,
            freezed=freezed,
            rename_duplicates=rename_duplicates,
            load_internal=load_internal,
        )



from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.types.datatype import DataType
from spinta.types.text.components import Text


@commands.load.register(Context, Text, dict, Manifest)
def load(context: Context, dtype: Text, data: dict, manifest: Manifest) -> DataType:
    props = {}
    for name, params in (dtype.langs or {}).items():
        place = dtype.prop.place + '@' + name
        prop = dtype.prop.__class__()
        prop.name = name
        prop.place = place
        prop.parent = dtype.prop
        prop.model = dtype.prop.model
        prop.list = dtype.prop.list
        commands.load(context, prop, params, manifest)
        dtype.prop.model.flatprops[place] = prop
        props[name] = prop
    dtype.langs = props
    return dtype



from spinta import commands
from spinta.components import Context
from spinta.manifests.rdf.components import RdfManifest


@commands.bootstrap.register(Context, RdfManifest)
def bootstrap(context: Context, manifest: RdfManifest):
    store = context.get('store')
    for backend in store.backends.values():
        commands.bootstrap(context, backend)



from __future__ import annotations

import csv
import pathlib
import logging
import textwrap
from operator import itemgetter
from itertools import zip_longest
from typing import Any
from typing import Callable
from typing import Dict
from typing import IO
from typing import Iterable
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast

import openpyxl
import xlsxwriter
from lark import ParseError
from tabulate import tabulate

from spinta import commands
from spinta import spyna
from spinta.backends import Backend
from spinta.backends.components import BackendOrigin
from spinta.components import Context, Base, PrepareGiven
from spinta.datasets.components import Resource
from spinta.dimensions.comments.components import Comment
from spinta.dimensions.enum.components import EnumItem
from spinta.components import Model
from spinta.components import Namespace
from spinta.components import Property
from spinta.core.enums import Access
from spinta.core.ufuncs import unparse
from spinta.datasets.components import Dataset
from spinta.dimensions.enum.components import Enums
from spinta.dimensions.lang.components import LangData
from spinta.dimensions.prefix.components import UriPrefix
from spinta.exceptions import MultipleErrors, InvalidBackRefReferenceAmount, DataTypeCannotBeUsedForNesting, \
    NestedDataTypeMissmatch
from spinta.exceptions import PropertyNotFound
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.tabular.components import ACCESS, URI
from spinta.manifests.tabular.components import BackendRow
from spinta.manifests.tabular.components import BaseRow
from spinta.manifests.tabular.components import CommentData
from spinta.manifests.tabular.components import DESCRIPTION
from spinta.manifests.tabular.components import DatasetRow
from spinta.manifests.tabular.components import ParamRow
from spinta.manifests.tabular.components import EnumRow
from spinta.manifests.tabular.components import ID
from spinta.manifests.tabular.components import MANIFEST_COLUMNS
from spinta.manifests.tabular.components import ManifestColumn
from spinta.manifests.tabular.components import ManifestRow
from spinta.manifests.tabular.components import ManifestTableRow
from spinta.manifests.tabular.components import ModelRow
from spinta.manifests.tabular.components import PREPARE
from spinta.manifests.tabular.components import PROPERTY
from spinta.manifests.tabular.components import PrefixRow
from spinta.manifests.tabular.components import PropertyRow
from spinta.manifests.tabular.components import REF
from spinta.manifests.tabular.components import ResourceRow
from spinta.manifests.tabular.components import SOURCE
from spinta.manifests.tabular.components import TITLE
from spinta.manifests.tabular.components import LEVEL
from spinta.manifests.tabular.components import TabularFormat
from spinta.manifests.tabular.constants import DATASET
from spinta.manifests.tabular.formats.gsheets import read_gsheets_manifest
from spinta.spyna import SpynaAST
from spinta.types.datatype import Ref, DataType, Denorm, Inherit, ExternalRef, BackRef, ArrayBackRef, Array, Object
from spinta.utils.data import take
from spinta.utils.schema import NA
from spinta.utils.schema import NotAvailable
from spinta.types.text.components import Text

log = logging.getLogger(__name__)

ParsedRow = Tuple[int, Dict[str, Any]]


MAIN_DIMENSIONS = [
    'dataset',
    'resource',
    'base',
    'model',
    'property',
]
EXTRA_DIMENSIONS = [
    '',
    'prefix',
    'enum',
    'param',
    'comment',
    'ns',
    'lang',
    'unique'
]


class TabularManifestError(Exception):
    pass


def _detect_header(
    path: Optional[str],
    line: int,  # Line number
    row: Iterable[str],
) -> List[str]:
    header = [h.strip().lower() for h in row]
    unknown_columns = set(header[:len(DATASET)]) - set(DATASET)
    if unknown_columns:
        unknown_columns = ', '.join(sorted(unknown_columns, key=header.index))
        raise TabularManifestError(
            f"{path}:{line}: Unknown columns: {unknown_columns}."
        )
    return header


def _detect_dimension(
    path: Optional[pathlib.Path],
    line: str,  # Line number with a prefix (depends on manifest format)
    row: Dict[str, str],
) -> Optional[str]:
    dimensions = [k for k in MAIN_DIMENSIONS if row[k]]

    if len(dimensions) == 1:
        return dimensions[0]

    if len(dimensions) > 1:
        dimensions = ', '.join([f'{k}: {row[k]}' for k in dimensions])
        raise TabularManifestError(
            f"{path}:{line}: In one row only single dimension can be used, "
            f"but found more than one: {dimensions}"
        )

    if row['type']:
        if row['type'] not in EXTRA_DIMENSIONS:
            raise TabularManifestError(
                f"{path}:{line}:type: Unknown additional dimension name "
                f"{row['type']}."
            )
        return row['type']

    return ''


def _parse_spyna(
    reader: TabularReader,
    formula: str,
) -> Union[SpynaAST, NotAvailable, None]:
    if formula:
        try:
            return spyna.parse(formula)
        except ParseError as e:
            reader.error(f"Error while parsing formula {formula!r}:\n{e}")
    return NA


class TabularReader:
    state: State
    path: str
    line: str
    type: str
    name: str
    data: ManifestRow               # Used when `appendable` is False
    rows: List[Dict[str, Any]]      # Used when `appendable` is True
    appendable: bool = False        # Tells if reader is appendable.

    def __init__(
        self,
        state: State,
        path: str,
        line: str,
    ):
        self.state = state
        self.path = path
        self.line = line
        self.data = {}
        self.rows = []

    def __str__(self):
        return f"<{type(self).__name__} name={self.name!r}>"

    def read(self, row: Dict[str, str]) -> None:
        raise NotImplementedError

    def append(self, row: Dict[str, str]) -> None:
        if any(row.values()):
            self.error(
                f"Updates are not supported in context of {self.type!r}."
            )

    def release(self, reader: TabularReader = None) -> bool:
        raise NotImplementedError

    def items(self) -> Iterator[ParsedRow]:
        if self.appendable:
            for data in self.rows:
                yield self.line, data
        else:
            yield self.line, self.data

    def enter(self) -> None:
        raise NotImplementedError

    def leave(self) -> None:
        raise NotImplementedError

    def error(self, message: str) -> None:
        raise TabularManifestError(f"{self.path}:{self.line}: {message}")


class ManifestReader(TabularReader):
    type: str = 'manifest'
    datasets: Set[str]
    namespaces: Set[str]
    data: ManifestTableRow

    def read(self, row: ManifestRow) -> None:
        self.name = str(self.path)
        self.data = {
            'type': 'manifest',
        }

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None

    def enter(self) -> None:
        self.datasets = set()
        self.namespaces = set()
        self.state.manifest = self

    def leave(self) -> None:
        self.state.manifest = None


class DatasetReader(TabularReader):
    type: str = 'dataset'
    data: DatasetRow

    def read(self, row: Dict[str, str]) -> None:
        self.name = row['dataset']

        if self.name == '/':
            self.data = {}
        else:
            if row['dataset'] in self.state.manifest.datasets:
                self.error("Dataset already defined.")

            self.data = {
                'type': 'dataset',
                'id': row['id'],
                'name': row['dataset'],
                'source': row['source'],
                'level': row['level'],
                'access': row['access'],
                'title': row['title'],
                'description': row['description'],
                'resources': {},
            }

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None or isinstance(reader, (
            ManifestReader,
            DatasetReader,
        )) or (
            isinstance(reader, (ResourceReader, ModelReader)) and
            self.name == '/'
        )

    def enter(self) -> None:
        self.state.dataset = self

    def leave(self) -> None:
        self.state.dataset = None


class ResourceReader(TabularReader):
    type: str = 'resource'
    data: Union[BackendRow, ResourceRow]

    def read(self, row: Dict[str, str]) -> None:
        self.name = row['resource']

        if self.name == '/':
            self.data = {}
        else:
            if self.state.dataset is None:
                self.read_backend(row)
            else:
                self.read_resource(row)

    def read_backend(self, row: Dict[str, str]) -> None:
        # Backends will be loaded using
        # `spinta.manifests.helpers._load_manifest_backends`.

        if 'backends' not in self.state.manifest.data:
            self.state.manifest.data['backends'] = {}
        backends = self.state.manifest.data['backends']

        if self.name in backends:
            self.error(
                f"Backend {self.name!r} with the same name already defined."
            )

        self.data = {
            'type': row['type'],
            'name': self.name,
            'dsn': row['source'],
            'title': row['title'],
            'description': row['description'],
        }

        backends[self.name] = self.data

    def read_resource(self, row: Dict[str, str]) -> None:
        dataset = _get_state_obj(self.state.dataset)
        dataset = dataset.data if dataset else None

        if self.name in dataset['resources']:
            self.error("Resource with the same name already defined in ")

        self.data = {
            'type': row['type'],
            'backend': row['ref'],
            'external': row['source'],
            'prepare': _parse_spyna(self, row[PREPARE]),
            'level': row['level'],
            'access': row['access'],
            'title': row['title'],
            'description': row['description'],
        }

        dataset['resources'][self.name] = self.data

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None or isinstance(reader, (
            ManifestReader,
            DatasetReader,
            ResourceReader,
        )) or (isinstance(reader, ModelReader) and self.name == '/')

    def enter(self) -> None:
        self.state.resource = self

    def leave(self) -> None:
        self.state.resource = None


class BaseReader(TabularReader):
    type: str = 'base'
    data: BaseRow

    def read(self, row: Dict[str, str]) -> None:
        self.name = row['base']

        if self.name == '/':
            self.data = {}
        else:
            dataset = _get_state_obj(self.state.dataset)
            dataset = dataset.data if dataset else None

            self.data = {
                'name': self.name,
                'model': get_relative_model_name(dataset, row['base']),
                'pk': (
                    [x.strip() for x in row['ref'].split(',')]
                    if row['ref'] else []
                ),
                'level': row['level']
            }

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None or isinstance(reader, (
            ManifestReader,
            DatasetReader,
            ResourceReader,
            BaseReader,
        )) or (isinstance(reader, ModelReader) and self.name == '/')

    def enter(self) -> None:
        self.state.base = self

    def leave(self) -> None:
        self.state.base = None


class ModelReader(TabularReader):
    type: str = 'model'
    data: ModelRow

    def read(self, row: Dict[str, str]) -> None:
        dataset = _get_state_obj(self.state.dataset)
        resource = _get_state_obj(self.state.resource)
        base = _get_state_obj(self.state.base)
        name = get_relative_model_name(
            dataset.data if dataset else None,
            row['model'],
        )

        if self.state.rename_duplicates:
            dup = 1
            _name = name
            while _name in self.state.models:
                _name = f'{name}_{dup}'
                dup += 1
            name = _name
        elif name in self.state.models:
            self.error(f"Model {name!r} with the same name is already defined.")

        self.name = name

        self.data = {
            'type': 'model',
            'id': row['id'],
            'name': name,
            'base': {
                'name': base.name,
                'parent': base.data['model'],
                'pk': base.data['pk'],
                'level': base.data['level']
            } if base and base.data else None,
            'level': row['level'],
            'access': row['access'],
            'title': row['title'],
            'description': row['description'],
            'properties': {},
            'uri': row['uri'],
            'unique': [([x.strip() for x in row['ref'].split(',')])] if row['ref'] else [],
            'external': {
                'dataset': dataset.name if dataset else '',
                'resource': resource.name if dataset and resource else '',
                'pk': (
                    [x.strip() for x in row['ref'].split(',')]
                    if row['ref'] else []
                ),
                'name': row['source'],
                'prepare': _parse_spyna(self, row[PREPARE]),
            },
        }
        if resource and not dataset:
            self.data['backend'] = resource.name

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None or isinstance(reader, (
            ManifestReader,
            DatasetReader,
            ResourceReader,
            BaseReader,
            ModelReader,
        ))

    def enter(self) -> None:
        self.state.model = self
        self.state.models.add(self.name)

    def leave(self) -> None:
        self.state.model = None


def _parse_property_ref(ref: str) -> Tuple[str, List[str]]:
    if '[' in ref:
        ref = ref.rstrip(']')
        ref_model, ref_props = ref.split('[', 1)
        ref_props = [p.strip() for p in ref_props.split(',')]
    else:
        ref_model = ref
        ref_props = []
    return ref_model, ref_props


def _parse_dtype_string(dtype: str) -> dict:
    args = []
    error = None
    required = unique = False
    invalid_args = []

    if '(' in dtype:
        dtype, args = dtype.split('(', 1)
        args, additional_args = args.split(')', 1)
        args = args.strip().rstrip(')')
        args = [a.strip() for a in args.split(',')]
    else:
        if len(dtype.split(None, 1)) > 1:
            dtype, additional_args = dtype.split(None, 1)
        else:
            additional_args = ""

    if additional_args:
        for arg in additional_args.split(None):
            if arg == 'required':
                required = True
            elif arg == 'unique':
                unique = True
            else:
                invalid_args.append(arg)
        if invalid_args:
            error = f'Invalid type arguments: {", ".join(invalid_args)}.'

    return {
        'type': dtype,
        'type_args': args,
        'required': required,
        'unique': unique,
        'error': error,
    }


def _get_type_repr(dtype: [DataType, str]):
    if isinstance(dtype, DataType):
        args = ''
        required = ' required' if dtype.required else ''
        unique = ' unique' if dtype.unique else ''

        model = dtype.prop.model
        if model.external.unknown_primary_key is False:
            if len(model.external.pkeys) == 1 and dtype.prop in model.external.pkeys:
                unique = ''
        if dtype.type_args:
            args = ', '.join(dtype.type_args)
            args = f'({args})'
        dtype_name = dtype.name if not isinstance(dtype, (Denorm, Inherit, ExternalRef, ArrayBackRef)) else dtype.get_type_repr()
        return f'{dtype_name}{args}{required}{unique}'
    else:
        args = ''
        required = ' required' if 'required' in dtype else ''
        unique = ' unique' if 'unique' in dtype else ''
        additional_args = []
        if '(' in dtype:
            dtype, args = dtype.split('(', 1)
            args, additional_args = args.split(')', 1)
            args = args.strip().rstrip(')')
            args = [a.strip() for a in args.split(',')]
            args = ', '.join(args)
            args = f'({args})'
        else:
            if len(dtype.split(None, 1)) > 1:
                dtype, additional_args = dtype.split(None, 1)
            else:
                dtype = dtype.strip(' ')
        if additional_args:
            if [additional_arg for additional_arg in additional_args.split(' ') if additional_arg not in [
                required.strip(' '), unique.strip(' ')]]:
                raise TabularManifestError
        return f'{dtype}{args}{required}{unique}'


class PropertyReader(TabularReader):
    type: str = 'property'
    data: PropertyRow
    enums: Set[str]

    def read(self, row: Dict[str, str]) -> None:
        full_prop, parent_prop, prop_name = _get_parent_data(self, row, row['property'])
        prop_data = _handle_datatype(self, row)
        if prop_data:
            prop_name = _combine_parent_with_prop(prop_name, prop_data, parent_prop, full_prop)
        self.name = prop_name
        self.data = full_prop
        self.state.model.data['properties'][prop_name] = self.data

    def append(self, row: Dict[str, str]) -> None:
        if not row['property']:
            result = row['prepare']

            if row['source']:
                if result:
                    split = result.split('.')
                    formula = split[0]
                    split_formula = formula.split('(')
                    formatted = row["source"].replace('"', '\\"').replace("'", "\\'")
                    reconstructed = f'{split_formula[0]}("{formatted}", {"(".join(split_formula[1:])}'
                    split[0] = reconstructed
                    result = '.'.join(split)
                else:
                    result = row['source']
            if not result:
                return
            self._append_prepare(row, result)

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None or isinstance(reader, (
            ManifestReader,
            DatasetReader,
            ResourceReader,
            BaseReader,
            ModelReader,
            PropertyReader,
            UniqueReader
        ))

    def enter(self) -> None:
        self.state.prop = self

    def leave(self) -> None:
        self._parse_prepare()
        if 'external' in self.data:
            self.data['external']['prepare'] = self.data.pop('prepare') if 'prepare' in self.data else NA

        self.state.prop = None

    def _parse_prepare(self):
        if "prepare" in self.data:
            self.data["prepare"] = _parse_spyna(self, self.data["prepare"])

    def _append_prepare(self, row: Dict[str, str], prepare: str):
        if "prepare" in self.data:
            prep = self.data["prepare"]
            self.data["prepare"] = f'{prep}.{prepare}' if prep else prepare
        self.data['prepare_given'].append(
            PrepareGiven(
                appended=True,
                source=row['source'],
                prepare=row['prepare']
            )
        )


ALLOWED_PARTIAL_TYPES = ['object', 'partial', 'ref']
ALLOWED_ARRAY_TYPES = ['array', 'partial_array']


def _initial_normal_property_schema(given_name: str, dtype: dict, row: dict):
    return {
        'type': dtype['type'],
        'type_args': dtype['type_args'],
        'prepare': row.get(PREPARE),
        'level': row.get(LEVEL),
        'access': row.get(ACCESS),
        'uri': row.get(URI),
        'title': row.get(TITLE),
        'description': row.get(DESCRIPTION),
        'required': dtype['required'],
        'unique': dtype['unique'],
        'given_name': given_name,
        'prepare_given': [],
        'explicitly_given': True
    }


def _initial_array_property_schema(given_name: str, dtype: dict, row: dict):
    result = _initial_normal_property_schema(given_name, dtype, row)
    result['items'] = {}
    return result


def _initial_partial_property_schema(given_name: str, dtype: dict, row: dict):
    result = _initial_normal_property_schema(given_name, dtype, row)
    result['properties'] = {}
    return result


def _initial_backref_property_schema(given_name: str, dtype: dict, row: dict):
    result = _initial_normal_property_schema(given_name, dtype, row)
    if given_name.endswith('[]'):
        result['type'] = 'array_backref'
    return result


def _initial_text_property_schema(given_name: str, dtype: dict, row: dict):
    result = _initial_normal_property_schema(given_name, dtype, row)
    result['langs'] = {}
    return result


def _datatype_handler(reader: PropertyReader, row: dict, initial_data_loader: Callable[[str, dict, dict], dict]):
    given_name = row['property']
    reader.name = _clean_up_prop_name(row['property'].split('.')[-1])

    if reader.state.model is None:
        context = reader.state.stack[-1]
        reader.error(
            f"Property {reader.name!r} must be defined in a model context. "
            f"Now it is defined in {context.name!r} {context.type} context."
        )
    _check_if_property_already_set(reader, row, given_name)
    dtype = _get_type_repr(row['type'])
    dtype = _parse_dtype_string(dtype)
    if dtype['error']:
        reader.error(
            dtype['error']
        )

    if reader.state.base and not dtype['type']:
        dtype['type'] = 'inherit'
    elif '.' in given_name and not dtype['type']:
        dtype['type'] = 'denorm'

    new_data = initial_data_loader(given_name, dtype, row)
    dataset = reader.state.dataset.data if reader.state.dataset else None

    custom_data = new_data
    if row['prepare']:
        custom_data['prepare_given'].append(
            PrepareGiven(
                appended=False,
                source='',
                prepare=row['prepare']
            )
        )
    if row['ref']:
        if dtype['type'] in ('ref', 'backref', 'generic'):
            ref_model, ref_props = _parse_property_ref(row['ref'])
            custom_data['model'] = get_relative_model_name(dataset, ref_model)
            if dtype['type'] == 'backref':
                if len(ref_props) > 1:
                    raise InvalidBackRefReferenceAmount(backref=reader.name)
                if len(ref_props) == 1:
                    custom_data['refprop'] = ref_props[0]
            else:
                custom_data['refprops'] = ref_props
        else:
            # TODO: Detect if ref is a unit or an enum.
            custom_data['enum'] = row['ref']
    if dataset or row['source']:
        custom_data['external'] = {
            'name': row['source'],
        }

    return new_data


def _string_datatype_handler(reader: PropertyReader, row: dict):
    given_name = row['property']
    reader.name = _clean_up_prop_name(row['property'].split('.')[-1])

    if reader.state.model is None:
        context = reader.state.stack[-1]
        reader.error(
            f"Property {reader.name!r} must be defined in a model context. "
            f"Now it is defined in {context.name!r} {context.type} context."
        )
    existing_data = _check_if_property_already_set(reader, row, given_name)
    if row['type'] == 'text' and existing_data:
        reader.error(
            f"Property {reader.name!r} with the same name is already "
            f"defined for this {reader.state.model.name!r} model."
        )

    dtype = _get_type_repr(row['type'])
    dtype = _parse_dtype_string(dtype)
    if dtype['error']:
        reader.error(
            dtype['error']
        )

    new_data = _initial_normal_property_schema(given_name, dtype, row)
    dataset = reader.state.dataset.data if reader.state.dataset else None

    if row['prepare']:
        new_data['prepare_given'].append(
            PrepareGiven(
                appended=False,
                source='',
                prepare=row['prepare']
            )
        )
    if row['ref']:
        new_data['enum'] = row['ref']
    if dataset or row['source']:
        new_data['external'] = {
            'name': row['source'],
        }

    lang = None
    if '@' in given_name:
        lang = given_name.split('@', 1)[-1]
    should_return = True
    if existing_data:
        if existing_data['type'] == 'text':
            should_return = False
            if lang and lang in existing_data['langs']:
                reader.error(
                    f"Language {lang} has already been set for the {existing_data['given_name']} property."
                )
            elif not lang:
                existing_data['langs']['C'] = new_data
            else:
                existing_data['langs'][lang] = new_data
        else:
            reader.error(
                f"Language can only be added to Text type properties."
            )
    elif lang and not existing_data:
        copy = new_data.copy()
        new_data = _initial_text_property_schema(given_name, dtype, {
            'property': row['property'],
            'access': row['access']
        })
        new_data['type'] = 'text'
        new_data['explicitly_given'] = False
        new_data['langs'] = {
            lang: copy
        }
    if should_return:
        return new_data


def _text_datatype_handler(reader: PropertyReader, row: dict):
    given_name = row['property']
    reader.name = _clean_up_prop_name(row['property'].split('.')[-1])

    if reader.state.model is None:
        context = reader.state.stack[-1]
        reader.error(
            f"Property {reader.name!r} must be defined in a model context. "
            f"Now it is defined in {context.name!r} {context.type} context."
        )
    result = _check_if_property_already_set(reader, row, given_name)
    if not (result and result['explicitly_given'] is False and result['type'] == 'text' or not result):
        reader.error(
            f"Property {reader.name!r} with the same name is already "
            f"defined for this {reader.state.model.name!r} model."
        )
    dtype = _get_type_repr(row['type'])
    dtype = _parse_dtype_string(dtype)
    if dtype['error']:
        reader.error(
            dtype['error']
        )

    new_data = _initial_text_property_schema(given_name, dtype, row)
    dataset = reader.state.dataset.data if reader.state.dataset else None

    if row['prepare']:
        new_data['prepare_given'].append(
            PrepareGiven(
                appended=False,
                source='',
                prepare=row['prepare']
            )
        )
    if row['ref']:
        new_data['enum'] = row['ref']
    if dataset or row['source']:
        new_data['external'] = {
            'name': row['source'],
        }
    temp_data = _empty_property(_initial_normal_property_schema(given_name, dtype, {
        'property': row['property'],
        'access': row['access'],
    }))
    temp_data['type'] = 'string'
    temp_data['external'] = new_data['external']
    if result:
        new_data['langs'] = result['langs']
        if new_data['level'] and int(new_data['level']) <= 3:
            new_data['langs']['C'] = temp_data
            if new_data['external']:
                new_data['external'] = {}
        result.update(new_data)
        return result
    else:
        if new_data['level'] and int(new_data['level']) <= 3:
            new_data['langs'] = {
                'C': temp_data
            }
            if new_data['external']:
                new_data['external'] = {}
        return new_data


def _default_datatype_handler(reader: PropertyReader, row: dict):
    return _datatype_handler(reader, row, _initial_normal_property_schema)


def _array_datatype_handler(reader: PropertyReader, row: dict):
    return _datatype_handler(reader, row, _initial_array_property_schema)


def _partial_datatype_handler(reader: PropertyReader, row: dict):
    return _datatype_handler(reader, row, _initial_partial_property_schema)


def _backref_datatype_handler(reader: PropertyReader, row: dict):
    return _datatype_handler(reader, row, _initial_backref_property_schema)


def _handle_datatype(reader: PropertyReader, row: dict):
    if row['type'] in DATATYPE_HANDLERS:
        handler = DATATYPE_HANDLERS[row['type']]
    else:
        handler = DATATYPE_HANDLERS['_default']
    return handler(reader, row)


DATATYPE_HANDLERS = {
    "_default": _default_datatype_handler,
    "partial_array": _array_datatype_handler,
    "array": _array_datatype_handler,
    "object": _partial_datatype_handler,
    "partial": _partial_datatype_handler,
    "ref": _partial_datatype_handler,
    "backref": _backref_datatype_handler,
    "text": _text_datatype_handler,
    "string": _string_datatype_handler
}


def _get_root_prop(reader: PropertyReader, name: str):
    if name in reader.state.model.data['properties']:
        return reader.state.model.data['properties'][name]
    return None


def _clean_up_prop_name(name: str):
    return name.replace('[]', '').split('@', 1)[0]


def _combine_previous_data(prop: dict, existing_prop: dict):
    if 'properties' in existing_prop:
        prop['properties'] = existing_prop['properties']
    elif 'items' in existing_prop:
        prop['items'] = existing_prop['items']
    return prop


def _combine_parent_with_prop(prop_name: str, prop: dict, parent_prop: dict, full_prop: dict):
    return_name = _clean_up_prop_name(prop['given_name'])
    if parent_prop:
        if parent_prop['type'] in ALLOWED_PARTIAL_TYPES:
            if prop_name in parent_prop['properties']:
                prop = _combine_previous_data(prop, parent_prop['properties'][prop_name])
            parent_prop['properties'][prop_name] = prop
            return_name = _clean_up_prop_name(full_prop['given_name'].split('.')[0])
        elif parent_prop['type'] in ALLOWED_ARRAY_TYPES:
            # Array backref edge to replace parent instead of update array.items
            if prop['type'] == 'array_backref':
                parent_prop.clear()
                parent_prop.update(prop)
            else:
                if parent_prop['items']:
                    prop = _combine_previous_data(prop, parent_prop['items'])
                parent_prop['items'] = prop
            return_name = _clean_up_prop_name(full_prop['given_name'].split('.')[0])
        else:
            full_prop.clear()
            full_prop.update(prop)
    else:
        if full_prop:
            if prop['type'] in ALLOWED_ARRAY_TYPES:
                prop['items'] = full_prop['items']
            elif prop['type'] in ALLOWED_PARTIAL_TYPES and 'properties' in full_prop:
                prop['properties'] = full_prop['properties']
        full_prop.update(prop)
    return return_name


def _empty_property(data: dict):
    data['explicitly_given'] = False
    return data


def _get_parent_data_array(reader: PropertyReader, given_row: dict, full_name: str, current_parent: dict):
    name = full_name.split('.')[-1]
    count = name.count('[]')
    root_name = name.replace('[]', '')
    empty_array_row = torow(DATASET, {
        'property': full_name,
        'type': 'partial_array',
        'access': given_row['access'],
    })
    if not current_parent:
        current_parent.update(_empty_property(_array_datatype_handler(reader, empty_array_row)))
    adjustment = 1 if current_parent['type'] in ALLOWED_ARRAY_TYPES else 0
    for i in range(count - adjustment):
        if current_parent['type'] in ALLOWED_ARRAY_TYPES:
            if current_parent['items'] and current_parent['items']['type'] not in ALLOWED_ARRAY_TYPES:
                raise NestedDataTypeMissmatch(initial=current_parent['type'], required='array')
            elif not current_parent['items']:
                current_parent['items'].update(_empty_property(_array_datatype_handler(reader, empty_array_row)))
            current_parent = current_parent['items']
        elif current_parent['type'] in ALLOWED_PARTIAL_TYPES:
            if root_name in current_parent['properties'] and current_parent['properties'][root_name]['type'] not in ALLOWED_ARRAY_TYPES:
                raise NestedDataTypeMissmatch(initial=current_parent['type'], required='array')
            elif root_name not in current_parent['properties']:
                current_parent['properties'][root_name] = _empty_property(_array_datatype_handler(reader, empty_array_row))
            current_parent = current_parent['properties'][root_name]
    return current_parent


def _get_parent_data_partial(reader: PropertyReader, given_row: dict, full_name: str, current_parent: dict):
    empty_partial_row = torow(DATASET, {
        'property': full_name,
        'type': 'partial',
        'access': given_row['access'],
    })
    name = _clean_up_prop_name(full_name.split('.')[-1])
    if not current_parent:
        current_parent.update(_empty_property(_partial_datatype_handler(reader, empty_partial_row)))
    else:
        if current_parent['type'] in ALLOWED_ARRAY_TYPES:
            if current_parent['items'] and current_parent['items']['type'] not in ALLOWED_PARTIAL_TYPES:
                raise NestedDataTypeMissmatch(initial=current_parent['type'], required='partial')
            elif not current_parent['items']:
                current_parent['items'].update(_empty_property(_partial_datatype_handler(reader, empty_partial_row)))
            current_parent = current_parent['items']
        elif current_parent['type'] in ALLOWED_PARTIAL_TYPES:
            if name in current_parent['properties'] and current_parent['properties'][name]['type'] not in ALLOWED_PARTIAL_TYPES:
                raise NestedDataTypeMissmatch(initial=current_parent['type'], required='partial')
            elif name not in current_parent['properties']:
                current_parent['properties'][name] = _empty_property(_partial_datatype_handler(reader, empty_partial_row))
            current_parent = current_parent['properties'][name]
    return current_parent


def _get_parent_data(reader: PropertyReader, given_row: dict, name: str):
    split_props = name.split('.')
    full_prop = {}
    root = _get_root_prop(reader, _clean_up_prop_name(split_props[0]))
    if root:
        full_prop.update(root)

    current_parent = None
    count = len(split_props)
    prop_name = name
    full_name = []
    for i, prop in enumerate(split_props):
        full_name.append(prop)
        if '[]' in prop:
            current_parent = _get_parent_data_array(reader, given_row, '.'.join(full_name), current_parent or full_prop)
        if i + 1 != count:
            if not current_parent and full_prop:
                current_parent = full_prop
            else:
                current_parent = _get_parent_data_partial(reader, given_row, '.'.join(full_name), current_parent or full_prop)
        else:
            prop_name = _clean_up_prop_name(prop)
        if current_parent and (current_parent['type'] not in ALLOWED_PARTIAL_TYPES and current_parent['type'] not in ALLOWED_ARRAY_TYPES):
            raise DataTypeCannotBeUsedForNesting(dtype=current_parent['type'])
    return full_prop, current_parent, prop_name


def _check_if_property_already_set(reader: PropertyReader, given_row: dict, full_name: str):
    split = full_name.split('.')
    base_name = _clean_up_prop_name(split[0])
    base = {}
    if base_name in reader.state.model.data['properties']:
        base = reader.state.model.data['properties'][base_name]
        if base_name == full_name and base['type'] != 'partial' and base['type'] != 'partial_array' and (base['type'] == 'text' and given_row['type'] != 'string') and base['explicitly_given']:
            reader.error(
                f"Property {reader.name!r} with the same name is already "
                f"defined for this {reader.state.model.name!r} model."
            )
    if not base:
        return

    for i, prop in enumerate(split):
        if '[]' in prop:
            count = prop.count('[]') + 1
            prop = prop.replace('[]', '')

            for a_i in range(count):
                if not base:
                    return
                if base['type'] in ALLOWED_ARRAY_TYPES:
                    base = base['items']
                elif base['type'] in ALLOWED_PARTIAL_TYPES:
                    base = base['properties'].get(prop, None)
                else:
                    raise DataTypeCannotBeUsedForNesting(dtype=base['type'])
        else:
            if i > 0:
                if not base:
                    return
                if base['type'] in ALLOWED_PARTIAL_TYPES:
                    base = base['properties'].get(prop, None)
                elif base['type'] in ALLOWED_ARRAY_TYPES:
                    base = base['items']
                else:
                    raise DataTypeCannotBeUsedForNesting(dtype=base['type'])

    if base:
        if (base['type'] in ALLOWED_PARTIAL_TYPES and given_row['type'] not in ALLOWED_PARTIAL_TYPES) or (base['type'] in ALLOWED_ARRAY_TYPES and given_row['type'] not in ALLOWED_ARRAY_TYPES):
            raise DataTypeCannotBeUsedForNesting(dtype=given_row['type'])

    return base


class AppendReader(TabularReader):
    type: str = 'append'
    data: ManifestRow

    def read(self, row: ManifestRow) -> None:
        self.name = row[REF]
        self.data = row

    def release(self, reader: TabularReader = None) -> bool:
        return True

    def enter(self) -> None:
        pass

    def leave(self) -> None:
        self.state.stack[-1].append(self.data)


class PrefixReader(TabularReader):
    type: str = 'prefix'
    data: PrefixRow

    def read(self, row: Dict[str, str]) -> None:
        if not row['ref']:
            # `ref` is a required parameter.
            return

        self.name = row['ref']

        node = (
            self.state.prop or
            self.state.model or
            self.state.base or
            self.state.resource or
            self.state.dataset or
            self.state.manifest
        )

        if 'prefixes' not in node.data:
            node.data['prefixes'] = {}

        prefixes = node.data['prefixes']

        if self.name in prefixes:
            self.error(
                f"Prefix {self.name!r} with the same name is already "
                f"defined for this {node.name!r} {node.type}."
            )

        self.data = {
            'id': row['id'],
            'eid': f'{self.path}:{self.line}',
            'type': self.type,
            'name': self.name,
            'uri': row['uri'],
            'title': row['title'],
            'description': row['description'],
        }

        prefixes[self.name] = self.data

    def append(self, row: Dict[str, str]) -> None:
        self.read(row)

    def release(self, reader: TabularReader = None) -> bool:
        return not isinstance(reader, AppendReader)

    def enter(self) -> None:
        pass

    def leave(self) -> None:
        pass


class NamespaceReader(TabularReader):
    type: str = 'ns'
    appendable: bool = True

    def read(self, row: Dict[str, str]) -> None:
        if not row['ref']:
            # `ref` is a required parameter.
            return

        self.name = row['ref']

        manifest = self.state.manifest

        if self.name in manifest.namespaces:
            self.error(
                f"Namespace {self.name!r} with the same name is already "
                f"defined."
            )

        manifest.namespaces.add(self.name)

        self.rows.append({
            'id': row['id'],
            'type': self.type,
            'name': self.name,
            'title': row['title'],
            'description': row['description'],
        })

    def append(self, row: Dict[str, str]) -> None:
        self.read(row)

    def release(self, reader: TabularReader = None) -> bool:
        return not isinstance(reader, AppendReader)

    def enter(self) -> None:
        pass

    def leave(self) -> None:
        pass


class ParamReader(TabularReader):
    type: str = 'param'
    data: ParamRow
    name: str = None

    def _get_node(self) -> TabularReader:
        return (
            self.state.prop or
            self.state.model or
            self.state.base or
            self.state.resource or
            self.state.dataset or
            self.state.manifest
        )

    def _get_data(self, name: str, row: ManifestRow):
        return {
            'name': name,
            'source': [row[SOURCE]],
            'prepare': [_parse_spyna(self, row[PREPARE])],
            'title': row[TITLE],
            'description': row[DESCRIPTION],
        }

    def _get_and_append_data(self, old: Dict, row: ManifestRow):
        old["source"].append(row[SOURCE])
        old["prepare"].append(_parse_spyna(self, row[PREPARE]))

    def _ensure_params_list(self, node: TabularReader, name: str) -> None:
        if 'params' not in node.data:
            node.data['params'] = {}

    def _check_param_name(self, node: TabularReader, name: str) -> None:
        if 'params' in node.data and name in node.data['params']:
            self.error(
                f"Parameter {name!r} with the same name already defined!"
            )

    def read(self, row: ManifestRow) -> None:
        node = self._get_node()

        self.name = row[REF]
        if not self.name:
            self.error("Parameter must have a name.")

        self._check_param_name(node, self.name)
        self._ensure_params_list(node, self.name)

        self.data = self._get_data(self.name, row)
        node.data['params'][self.name] = self.data

    def append(self, row: ManifestRow) -> None:
        node = self._get_node()

        if row[REF]:
            self.name = row[REF]
            self._check_param_name(node, self.name)
            self._ensure_params_list(node, self.name)

        self._get_and_append_data(node.data['params'][self.name], row)

    def release(self, reader: TabularReader = None) -> bool:
        return not isinstance(reader, (AppendReader, LangReader))

    def enter(self) -> None:
        pass

    def leave(self) -> None:
        pass


class EnumReader(TabularReader):
    type: str = 'enum'
    data: EnumRow
    name: str = None

    def read(self, row: ManifestRow) -> None:
        if row[REF]:
            self.name = row[REF]
        else:
            self.name = self.name or ''

        if not any([
            row[SOURCE],
            row[PREPARE],
            row[ACCESS],
            row[TITLE],
            row[DESCRIPTION],
        ]):
            return

        # source = row[SOURCE] if row[SOURCE] is not None else row[PREPARE]
        source = str(row[SOURCE]) or row[PREPARE]
        if not source:
            self.error(
                "At least source or prepare must be specified for an enum."
            )

        if row[LEVEL]:
            self.error(f"Enum's do not have a level, but level {row[LEVEL]!r} is given.")

        self.data = {
            'name': self.name,
            'source': row[SOURCE],
            'prepare': _parse_spyna(self, row[PREPARE]),
            'access': row[ACCESS],
            'title': row[TITLE],
            'description': row[DESCRIPTION],
        }

        node = (
            self.state.prop or
            self.state.model or
            self.state.base or
            self.state.resource or
            self.state.dataset or
            self.state.manifest
        )

        if 'enums' not in node.data:
            node.data['enums'] = {}

        if self.name not in node.data['enums']:
            node.data['enums'][self.name] = {}

        enum = node.data['enums'][self.name]

        if source in enum:
            self.error(
                f"Enum {self.name!r} item {source!r} with the same value is "
                f"already defined."
            )
        enum[source] = self.data

    def append(self, row: ManifestRow) -> None:
        self.read(row)

    def release(self, reader: TabularReader = None) -> bool:
        return not isinstance(reader, (AppendReader, LangReader))

    def enter(self) -> None:
        pass

    def leave(self) -> None:
        pass


class LangReader(TabularReader):
    type: str = 'lang'

    def read(self, row: ManifestRow) -> None:
        reader = self.state.stack[-1]
        if not isinstance(reader, (
            DatasetReader,
            ResourceReader,
            BaseReader,
            ModelReader,
            PropertyReader,
            EnumReader,
        )):
            self.error(f'Language metadata is not supported on {reader.type}.')
            return

        if 'lang' not in reader.data:
            reader.data['lang'] = {}

        lang = reader.data['lang']

        self.name = row[REF]

        if self.name in lang:
            self.error(
                f"Language {self.name!r} with the same name is already "
                f"defined for this {reader.name!r} {reader.type}."
            )

        lang[self.name] = {
            'id': row[ID],
            'eid': f'{self.path}:{self.line}',
            'type': self.type,
            'ref': self.name,
            'title': row[TITLE],
            'description': row[DESCRIPTION],
        }

    def append(self, row: ManifestRow) -> None:
        self.read(row)

    def release(self, reader: TabularReader = None) -> bool:
        return not isinstance(reader, AppendReader)

    def enter(self) -> None:
        pass

    def leave(self) -> None:
        pass


class UniqueReader(TabularReader):
    type: str = 'unique'

    def read(self, row: ManifestRow) -> None:
        self.name = row[REF]
        reader = self.state.stack[-1]

        if not isinstance(reader, (
            ModelReader,
            UniqueReader,
            AppendReader
        )):
            self.error(f'Unique reader is not supported for {reader.type}.')
            return

        if self.type not in reader.data:
            reader.data['unique'] = []

    def append(self, row: ManifestRow) -> None:
        self.read(row)

    def release(self, reader: TabularReader = None) -> bool:
        return not isinstance(reader, AppendReader)

    def enter(self) -> None:
        data = [row.strip() for row in self.name.split(',')]
        if data not in self.state.model.data['unique']:
            self.state.model.data['unique'].append(data)

    def leave(self) -> None:
        pass


class CommentReader(TabularReader):
    type: str = 'comment'
    data: CommentData

    def read(self, row: ManifestRow) -> None:
        reader = self.state.stack[-1]

        if 'comments' not in reader.data:
            reader.data['comments'] = []

        comments = reader.data['comments']

        comments.append({
            'id': row[ID],
            'parent': row[REF],
            'author': row[SOURCE],
            'access': row[ACCESS],
            # TODO: parse datetime
            'created': row[TITLE],
            'comment': row[DESCRIPTION],
        })

    def append(self, row: ManifestRow) -> None:
        self.read(row)

    def release(self, reader: TabularReader = None) -> bool:
        return not isinstance(reader, AppendReader)

    def enter(self) -> None:
        pass

    def leave(self) -> None:
        pass


READERS = {
    # Main dimensions
    'dataset': DatasetReader,
    'resource': ResourceReader,
    'base': BaseReader,
    'model': ModelReader,
    'property': PropertyReader,

    # Extra dimensions
    '': AppendReader,
    'prefix': PrefixReader,
    'ns': NamespaceReader,
    'param': ParamReader,
    'enum': EnumReader,
    'lang': LangReader,
    'comment': CommentReader,
    'unique': UniqueReader
}


class State:
    stack: List[TabularReader]

    backends: Dict[str, Dict[str, str]] = None

    models: Set[str]

    manifest: ManifestReader = None
    dataset: DatasetReader = None
    resource: ResourceReader = None
    base: BaseReader = None
    model: ModelReader = None
    prop: PropertyReader = None

    rename_duplicates: bool = False

    def __init__(self):
        self.stack = []
        self.models = set()

    def release(self, reader: TabularReader = None) -> Iterator[ParsedRow]:
        for parent in list(reversed(self.stack)):
            if parent.release(reader):
                if isinstance(parent, (
                    ManifestReader,
                    NamespaceReader,
                    DatasetReader,
                    ModelReader,
                )) and parent.name != "/":
                    yield from parent.items()
                self.stack.pop()
                parent.leave()
            else:
                break

        if reader:
            reader.enter()
            self.stack.append(reader)


def _read_tabular_manifest_rows(
    path: Optional[str],
    rows: Iterator[Tuple[str, List[str]]],
    *,
    rename_duplicates: bool = True,
) -> Iterator[ParsedRow]:
    _, header = next(rows, (None, None))
    if header is None:
        # Looks like an empty file.
        return
    header = _detect_header(path, 1, header)

    defaults = {k: '' for k in MANIFEST_COLUMNS}

    state = State()
    state.rename_duplicates = rename_duplicates
    reader = ManifestReader(state, path, '1')
    reader.read({})
    yield from state.release(reader)

    for line, row in rows:
        _check_row_size(path, line, header, row)
        row = dict(zip(header, row))
        row = {**defaults, **row}
        dimension = _detect_dimension(path, line, row)
        Reader = READERS[dimension]
        reader = Reader(state, path, line)
        reader.read(row)
        yield from state.release(reader)

    yield from state.release()


def _check_row_size(
    path: Optional[str],
    line: int,
    header: List[str],
    row: List[str],
):
    if len(header) != len(row):
        header = ['header', 'row']
        table = [
            ['∅' if x is None else x for x in v]
            for v in zip_longest(header, row)
        ]
        table = tabulate(table, headers=['header', 'row'])
        raise TabularManifestError(
            f"{path}:{line}: "
            "Number of row cells do not match table header, see what is "
            "missing, missing cells marked with ∅ symbol:\n"
            f"{table}"
        )


def read_tabular_manifest(
    format_: TabularFormat = None,
    *,
    path: str = None,
    file: IO = None,
    rename_duplicates: bool = False,
) -> Iterator[ParsedRow]:
    if format_ == TabularFormat.GSHEETS:
        rows = read_gsheets_manifest(path)
    elif format_ == TabularFormat.CSV:
        rows = _read_csv_manifest(path, file)
    elif format_ == TabularFormat.ASCII:
        rows = _read_txt_manifest(path, file)
    elif format_ == TabularFormat.XLSX:
        rows = _read_xlsx_manifest(path)
    else:
        raise ValueError(f"Unknown tabular manifest format {format_!r}.")

    yield from _read_tabular_manifest_rows(
        path,
        rows,
        rename_duplicates=rename_duplicates,
    )


def _read_txt_manifest(
    path: str,
    file: IO[str] = None,
) -> Iterator[Tuple[str, List[str]]]:
    if file:
        yield from _read_ascii_tabular_manifest(file)
    else:
        with pathlib.Path(path).open(encoding='utf-8-sig') as f:
            yield from _read_ascii_tabular_manifest(f)


def _read_csv_manifest(
    path: str,
    file: IO[str] = None,
) -> Iterator[Tuple[str, List[str]]]:
    if file:
        rows = csv.reader(file)
        for i, row in enumerate(rows, 1):
            yield str(i), row
    else:
        with pathlib.Path(path).open(encoding='utf-8') as f:
            rows = csv.reader(f)
            for i, row in enumerate(rows, 1):
                yield str(i), row


def _empty_rows_counter():
    empty_rows = 0

    def _counter(row):
        nonlocal empty_rows
        if any(row):
            empty_rows = 0
        else:
            empty_rows += 1
        return empty_rows

    return _counter


def _read_xlsx_manifest(path: str) -> Iterator[Tuple[str, List[str]]]:
    wb = openpyxl.load_workbook(path)

    yield '1', DATASET

    for sheet in wb:
        rows = sheet.iter_rows(values_only=True)
        cols = next(rows, None)
        if cols is None:
            continue
        cols = normalizes_columns(cols)
        cols = [cols.index(c) if c in cols else None for c in DATASET]

        empty_rows = _empty_rows_counter()
        for i, row in enumerate(rows, 2):
            row = [row[c] if c is not None else None for c in cols]
            yield f'{sheet.title}:{i}', row

            if empty_rows(row) > 100:
                log.warning(
                    f"Too many consequent empty rows, stop reading {path} "
                    f"at {i} row."
                )
                break


def striptable(table):
    return textwrap.dedent(table).strip()


def _join_escapes(row: List[str]) -> List[str]:
    res = []
    for v in row:
        if res and res[-1] and res[-1].endswith('\\'):
            res[-1] = res[-1][:-1] + '|' + v
        else:
            res.append(v)
    return res


def _read_ascii_tabular_manifest(
    lines: Iterable[str],
    *,
    check_column_names: bool = True,
) -> Iterator[Tuple[str, List[str]]]:
    lines = (line.strip() for line in lines)
    lines = filter(None, lines)

    # Read header
    header = next(lines, None)
    if header is None:
        return
    header = normalizes_columns(
        header.split('|'),
        check_column_names=check_column_names,
    )
    yield '1', header

    # Find index where dimension columns end.
    dim = sum(1 for h in header if h in DATASET[:6])
    for i, line in enumerate(lines, 2):
        row = _join_escapes(line.split('|'))
        row = [x.strip() for x in row]
        row = row[:len(header)]
        rem = len(header) - len(row)
        row = row[:dim - rem] + [''] * rem + row[dim - rem:]
        assert len(header) == len(row), line
        yield str(i), row


def read_ascii_tabular_rows(
    manifest: str,
    *,
    strip: bool = False,
    check_column_names: bool = True,
) -> Iterator[List[str]]:
    if strip:
        manifest = striptable(manifest)
    rows = _read_ascii_tabular_manifest(
        manifest.splitlines(),
        check_column_names=check_column_names,
    )
    for line, row in rows:
        yield row


def read_ascii_tabular_manifest(
    manifest: str,
    *,
    strip: bool = False,
    rename_duplicates: bool = False,
) -> Iterator[ParsedRow]:
    if strip:
        manifest = striptable(manifest)
    rows = _read_ascii_tabular_manifest(manifest.splitlines())
    yield from _read_tabular_manifest_rows(
        None,
        rows,
        rename_duplicates=rename_duplicates,
    )


def load_ascii_tabular_manifest(
    context: Context,
    manifest: Manifest,
    manifest_ascii_table: str,
    *,
    strip: bool = False,
) -> None:
    schemas = read_ascii_tabular_manifest(manifest_ascii_table, strip=strip)
    load_manifest_nodes(context, manifest, schemas)
    commands.link(context, manifest)


def get_relative_model_name(dataset: [str, dict], name: str) -> str:
    if isinstance(dataset, str):
        return name.replace(dataset, '')
    if name.startswith('/'):
        return name[1:]
    elif dataset is None:
        return name
    else:
        return '/'.join([
            dataset['name'],
            name,
        ])


def to_relative_model_name(model: Model, dataset: Dataset = None) -> str:
    """Convert absolute model `name` to relative."""
    if dataset is None:
        return model.name
    if model.name.startswith(dataset.name):
        prefix = dataset.name
        return model.name[len(prefix) + 1:]
    else:
        return '/' + model.name


def tabular_eid(model: Model):
    if isinstance(model.eid, int):
        return model.eid
    else:
        return 0


class OrderBy(NamedTuple):
    func: Callable[[Union[Dataset, Model, Property, EnumItem]], Any]
    reverse: bool = False


def _order_datasets_by_access(dataset: Dataset):
    return dataset.access or Access.private


def _order_datasets_by_name(dataset: Dataset):
    return dataset.name


DATASETS_ORDER_BY = {
    'access': OrderBy(_order_datasets_by_access, reverse=True),
    'default': OrderBy(_order_datasets_by_name),
}


def _order_models_by_access(model: Model):
    return model.access or Access.private


MODELS_ORDER_BY = {
    'access': OrderBy(_order_models_by_access, reverse=True),
    'default': OrderBy(tabular_eid),
}


def _order_properties_by_access(prop: Property):
    return prop.access or Access.private


PROPERTIES_ORDER_BY = {
    'access': OrderBy(_order_properties_by_access, reverse=True),
}


T = TypeVar('T', Dataset, Model, Property, EnumItem)


def sort(
    ordering: Dict[str, OrderBy],
    items: Iterable[T],
    order_by: Optional[str],
) -> Iterable[T]:
    order: Optional[OrderBy] = None

    if order_by:
        order = ordering[order_by]
    elif 'default' in ordering:
        order = ordering['default']

    if order:
        return sorted(items, key=order.func, reverse=order.reverse)
    else:
        return items


def _end_marker(name):
    yield torow(DATASET, {
        name: "/"
    })


def _prefixes_to_tabular(
    prefixes: Dict[str, UriPrefix],
    *,
    separator: bool = False,
) -> Iterator[ManifestRow]:
    first = True
    for name, prefix in prefixes.items():
        yield torow(DATASET, {
            'id': prefix.id,
            'type': prefix.type if first else '',
            'ref': name,
            'uri': prefix.uri,
            'title': prefix.title,
            'description': prefix.description,
        })
        first = False

    if separator and prefixes:
        yield torow(DATASET, {})


def _backends_to_tabular(
    backends: Dict[str, Backend],
    *,
    separator: bool = False,
) -> Iterator[ManifestRow]:
    for name, backend in backends.items():
        yield torow(DATASET, {
            'type': backend.type,
            'resource': name,
            'source': backend.config.get('dsn'),
        })

    if separator and backends:
        yield torow(DATASET, {})


def _namespaces_to_tabular(
    namespaces: Dict[str, Namespace],
    *,
    separator: bool = False,
) -> Iterator[ManifestRow]:
    namespaces = {
        k: ns
        for k, ns in namespaces.items() if not ns.generated
    }
    first = True
    for name, ns in namespaces.items():
        yield torow(DATASET, {
            'type': ns.type if first else '',
            'ref': name,
            'title': ns.title,
            'description': ns.description,
        })
        first = False

    if separator and namespaces:
        yield torow(DATASET, {})


def _order_enums_by_access(item: EnumItem):
    return item.access or Access.private


ENUMS_ORDER_BY = {
    'access': OrderBy(_order_enums_by_access, reverse=True),
}


def _enums_to_tabular(
    enums: Optional[Enums],
    *,
    external: bool = True,
    access: Access = Access.private,
    order_by: ManifestColumn = None,
    separator: bool = False,
) -> Iterator[ManifestRow]:
    if enums is None:
        return
    for name, enum in enums.items():
        first = True
        items = sort(ENUMS_ORDER_BY, enum.values(), order_by)
        for item in items:
            if item.access is not None and item.access < access:
                continue
            yield torow(DATASET, {
                'type': 'enum' if first else '',
                'ref': name if first else '',
                'source': item.source if external else '',
                'prepare': unparse(item.prepare),
                'access': item.given.access,
                'title': item.title,
                'description': item.description,
            })
            if lang := list(_lang_to_tabular(item.lang)):
                first = True
                yield from lang
            else:
                first = False

    if separator and enums:
        yield torow(DATASET, {})


def _lang_to_tabular(
    lang: Optional[LangData],
) -> Iterator[ManifestRow]:
    if lang is None:
        return
    first = True
    for name, data in sorted(lang.items(), key=itemgetter(0)):
        yield torow(DATASET, {
            'type': 'lang' if first else '',
            'ref': name if first else '',
            'title': data['title'],
            'description': data['description'],
        })
        first = False


def _text_to_tabular(
    prop
):
    if not isinstance(prop.dtype, Text):
        return
    for lang in prop.dtype.langs:
        yield torow(DATASET, {
            'property': prop.name + '@' + lang,
            'type': prop.dtype.name,
            'level': prop.level.value if prop.level is not None else '',
            'access': prop.given.access
        })


def _comments_to_tabular(
    comments: Optional[List[Comment]],
    *,
    access: Access = Access.private,
) -> Iterator[ManifestRow]:
    if comments is None:
        return
    first = True
    for comment in comments:
        if comment.access < access:
            return
        yield torow(DATASET, {
            'id': comment.id,
            'type': 'comment' if first else '',
            'ref': comment.parent,
            'source': comment.author,
            'access': comment.given.access,
            'title': comment.created,
            'description': comment.comment,
        })
        first = False


def _unique_to_tabular(model_unique_data, hide_list: List) -> Iterator[ManifestRow]:
    if not model_unique_data:
        return
    for row in model_unique_data:
        if row not in hide_list:
            yield torow(DATASET, {
                'type': 'unique',
                'ref': ', '.join([r.name for r in row])
            })


def _params_to_tabular(params_data) -> Iterator[ManifestRow]:
    if not params_data:
        return
    for param, values in params_data.items():
        for i in range(len(values["source"])):
            if isinstance(values["prepare"][i], NotAvailable):
                prepare = ''
            else:
                prepare = spyna.unparse(values["prepare"][i])
            if not (isinstance(values["prepare"][i], NotAvailable) and values['source'][i] == ''):
                if i == 0:
                    yield torow(DATASET, {
                        'type': 'param',
                        'ref': param,
                        'source': values["source"][i],
                        'prepare': prepare,
                        'title': values["title"],
                        'description': values["description"]
                    })
                else:
                    yield torow(DATASET, {
                        'source': values["source"][i],
                        'prepare': prepare
                    })


def _dataset_to_tabular(
    dataset: Dataset,
    *,
    external: bool = True,
    access: Access = Access.private,
    order_by: ManifestColumn = None,
) -> Iterator[ManifestRow]:
    yield torow(DATASET, {
        'id': dataset.id,
        'dataset': dataset.name,
        'level': dataset.level,
        'access': dataset.given.access,
        'title': dataset.title,
        'description': dataset.description,
    })
    yield from _lang_to_tabular(dataset.lang)
    yield from _prefixes_to_tabular(dataset.prefixes, separator=True)
    yield from _enums_to_tabular(
        dataset.ns.enums,
        external=external,
        access=access,
        order_by=order_by,
    )


def _resource_to_tabular(
    resource: Resource,
    *,
    external: bool = True,
    access: Access = Access.private,
) -> Iterator[ManifestRow]:
    backend = resource.backend
    yield torow(DATASET, {
        'resource': resource.name,
        'source': resource.external if external else '',
        'prepare': unparse(resource.prepare or NA) if external else '',
        'type': resource.type,
        'ref': (
            backend.name
            if (
                external and
                backend and
                backend.origin != BackendOrigin.resource
            )
            else ''
        ),
        'level': resource.level,
        'access': resource.given.access,
        'title': resource.title,
        'description': resource.description,
    })
    yield from _params_to_tabular(resource.params)
    yield from _comments_to_tabular(resource.comments, access=access)
    yield from _lang_to_tabular(resource.lang)


def _base_to_tabular(
    base: Base,
) -> Iterator[ManifestRow]:
    data = {
        'base': base.name,
        'level': base.level.value if base.level else "",
    }
    if base.pk:
        data['ref'] = ', '.join([pk.place for pk in base.pk])
    yield torow(DATASET, data)
    yield from _lang_to_tabular(base.lang)


def _property_to_tabular(
    prop: Property,
    *,
    external: bool = True,
    access: Access = Access.private,
    order_by: ManifestColumn = None,
) -> Iterator[ManifestRow]:
    if prop.name.startswith('_'):
        return

    if prop.access is not None and prop.access < access:
        return

    data = {
        'property': prop.given.name,
        'type': _get_type_repr(prop.dtype),
        'level': prop.level.value if prop.level else "",
        'access': prop.given.access,
        'uri': prop.uri,
        'title': prop.title,
        'description': prop.description,
    }

    if external and prop.external:
        if isinstance(prop.external, list):
            # data['source'] = ', '.join(x.name for x in prop.external)
            # data['prepare'] = ', '.join(
            #     unparse(x.prepare or NA)
            #     for x in prop.external if x.prepare
            # )
            raise DeprecationWarning(
                "Source can't be a list, use prepare instead."
            )
        elif prop.external:
            data['source'] = prop.external.name
            data['prepare'] = unparse(prop.external.prepare or NA)
    yield_rows = []
    if isinstance(prop.dtype, Array):
        yield_array_row = prop.dtype.items
        yield_rows.append(yield_array_row)
    elif isinstance(prop.dtype, Ref):
        model = prop.model
        if model.external and model.external.dataset:
            data['ref'] = to_relative_model_name(
                prop.dtype.model,
                model.external.dataset,
            )
            pkeys = prop.dtype.model.external.pkeys
            rkeys = prop.dtype.refprops
            if rkeys and pkeys != rkeys:
                rkeys = ', '.join([p.place for p in rkeys])
                data['ref'] += f'[{rkeys}]'
        else:
            data['ref'] = prop.dtype.model.name

        if prop.dtype.properties:
            for obj_prop in prop.dtype.properties.values():
                yield_rows.append(obj_prop)
    elif isinstance(prop.dtype, BackRef):
        model = prop.model
        if model.external and model.external.dataset:
            data['ref'] = to_relative_model_name(
                prop.dtype.model,
                model.external.dataset,
            )
            rkey = prop.dtype.refprop.place
            if prop.dtype.explicit:
                data['ref'] += f'[{rkey}]'
        else:
            data['ref'] = prop.dtype.model.name

        for denorm_prop in prop.dtype.properties.values():
            yield_rows.append(denorm_prop)
    elif isinstance(prop.dtype, Object):
        for obj_prop in prop.dtype.properties.values():
            yield_rows.append(obj_prop)
    elif isinstance(prop.dtype, Text):
        for lang_prop in prop.dtype.langs.values():
            yield_rows.append(lang_prop)

    elif prop.enum is not None:
        data['ref'] = prop.given.enum
    elif prop.unit is not None:
        data['ref'] = prop.given.unit
    data, prepare_rows = _prepare_to_tabular(data, prop)
    if prop.given.explicit:
        yield torow(DATASET, data)
    yield from prepare_rows
    yield from _comments_to_tabular(prop.comments, access=access)
    yield from _lang_to_tabular(prop.lang)
    yield from _enums_to_tabular(
        prop.enums,
        external=external,
        access=access,
        order_by=order_by,
    )
    if yield_rows:
        for yield_row in yield_rows:
            if yield_row:
                yield from _property_to_tabular(yield_row, external=external, access=access, order_by=order_by)


def _prepare_to_tabular(data, prop):
    prep_rows = []
    if prop.given.prepare:
        data['prepare'] = ''
        for prep in prop.given.prepare:
            if prep['appended']:
                prep_rows.append(torow(DATASET, {
                    'source': prep['source'],
                    'prepare': prep['prepare']
                }))
            else:
                if prop.external:
                    data['prepare'] = prep['prepare']
    return data, prep_rows


def _model_to_tabular(
    model: Model,
    *,
    external: bool = True,
    access: Access = Access.private,
    order_by: ManifestColumn = None,
) -> Iterator[ManifestRow]:
    data = {
        'id': model.id,
        'model': model.name,
        'level': model.level.value if model.level else "",
        'access': model.given.access,
        'title': model.title,
        'description': model.description,
        'uri': model.uri if model.uri else "",
    }
    if model.external and model.external.dataset:
        data['model'] = to_relative_model_name(
            model,
            model.external.dataset,
        )
    if external and model.external:
        data.update({
            'source': model.external.name,
            'prepare': unparse(model.external.prepare or NA),
        })
        if (
            not model.external.unknown_primary_key and
            all(p.access >= access for p in model.external.pkeys)
        ):
            # Add `ref` only if all properties are available in the
            # resulting manifest.
            data['ref'] = ', '.join([
                p.name for p in model.external.pkeys
            ])

    hide_list = []
    if model.external:
        if not model.external.unknown_primary_key:
            hide_list = [model.external.pkeys]
    yield torow(DATASET, data)
    yield from _params_to_tabular(model.params)
    yield from _comments_to_tabular(model.comments, access=access)
    yield from _lang_to_tabular(model.lang)
    yield from _unique_to_tabular(model.unique, hide_list)
    props = sort(PROPERTIES_ORDER_BY, model.properties.values(), order_by)
    for prop in props:
        yield from _property_to_tabular(
            prop,
            external=external,
            access=access,
            order_by=order_by,
        )


def datasets_to_tabular(
    manifest: Manifest,
    *,
    external: bool = True,   # clean content of source and prepare
    access: Access = Access.private,
    internal: bool = False,  # internal models with _ prefix like _txn
    order_by: ManifestColumn = None,
) -> Iterator[ManifestRow]:
    yield from _prefixes_to_tabular(manifest.prefixes, separator=True)
    yield from _backends_to_tabular(manifest.backends, separator=True)
    yield from _namespaces_to_tabular(manifest.namespaces, separator=True)
    yield from _enums_to_tabular(
        manifest.enums,
        external=external,
        access=access,
        order_by=order_by,
        separator=True,
    )

    seen_datasets = set()
    dataset = None
    resource = None
    base = None
    models = manifest.models if internal else take(manifest.models)
    models = sort(MODELS_ORDER_BY, models.values(), order_by)

    separator = False
    for model in models:
        if model.access < access:
            continue

        if model.external:
            if dataset is None or (model.external.dataset and dataset.name != model.external.dataset.name):
                dataset = model.external.dataset
                if dataset:
                    seen_datasets.add(dataset.name)
                    resource = None
                    separator = True
                    yield from _dataset_to_tabular(
                        dataset,
                        external=external,
                        access=access,
                        order_by=order_by,
                    )
            elif dataset is not None and \
                    model.external.dataset is None:
                dataset = None
                resource = None
                base = None
                separator = True
                yield from _end_marker('dataset')

            if external and model.external and model.external.resource and (
                resource is None or
                resource.name != model.external.resource.name
            ):
                resource = model.external.resource
                if resource:
                    separator = True
                    yield from _resource_to_tabular(
                        resource,
                        external=external,
                        access=access,
                    )
            elif external and \
                    model.external and \
                    model.external.resource is None and \
                    dataset is not None and \
                    resource is not None:
                base = None
                yield from _end_marker('resource')

        if separator:
            yield torow(DATASET, {})
        else:
            separator = False

        if model.base and (not base or not is_base_same(model.base, base)):
            base = model.base
            yield from _base_to_tabular(model.base)
        elif base and not model.base:
            base = None
            yield from _end_marker("base")
        yield from _model_to_tabular(
            model,
            external=external,
            access=access,
            order_by=order_by,
        )

    datasets = sort(DATASETS_ORDER_BY, manifest.datasets.values(), order_by)
    for dataset in datasets:
        if dataset.name in seen_datasets:
            continue
        yield from _dataset_to_tabular(
            dataset,
            external=external,
            access=access,
            order_by=order_by,
        )
        for resource in dataset.resources.values():
            yield from _resource_to_tabular(resource)


def is_base_same(a: Base, b: Base):
    return a.name == b.name and a.level == b.level and a.pk == b.pk


def torow(keys, values) -> ManifestRow:
    return {k: values.get(k) for k in keys}


def render_tabular_manifest(
    manifest: Manifest,
    cols: List[ManifestColumn] = None,
    *,
    sizes: Dict[ManifestColumn, int] = None,
) -> str:
    rows = datasets_to_tabular(manifest)
    return render_tabular_manifest_rows(rows, cols, sizes=sizes)


def render_tabular_manifest_rows(
    rows: Iterable[ManifestRow],
    cols: List[ManifestColumn] = None,
    *,
    sizes: Dict[ManifestColumn, int] = None,
) -> str:
    cols = cols or MANIFEST_COLUMNS
    hs = 1 if ID in cols else 0  # hierarchical cols start
    he = cols.index(PROPERTY)    # hierarchical cols end
    hsize = 1                      # hierarchical column size
    bsize = 3                      # border size
    if sizes is None:
        sizes = dict(
            [(c, len(c)) for c in cols[:hs]] +
            [(c, 1) for c in cols[hs:he]] +
            [(c, len(c)) for c in cols[he:]]
        )
        rows = list(rows)
        for row in rows:
            for i, col in enumerate(cols):
                val = '' if row[col] is None else str(row[col])
                if col == ID:
                    sizes[col] = 2
                elif i < he:
                    size = (hsize + bsize) * (he - hs - i) + sizes[PROPERTY]
                    if size < len(val):
                        sizes[PROPERTY] += len(val) - size
                elif sizes[col] < len(val):
                    sizes[col] = len(val)

    line = []
    for col in cols:
        size = sizes[col]
        line.append(col[:size].ljust(size))
    lines = [line]

    for row in rows:
        if ID in cols:
            line = [row[ID][:2] if row[ID] else '  ']
        else:
            line = []

        for i, col in enumerate(cols[hs:he + 1]):
            val = row[col] or ''
            if val:
                depth = i
                break
        else:
            val = ''
            depth = 0

        line += [' ' * hsize] * depth
        size = (hsize + bsize) * (he - hs - depth) + sizes[PROPERTY]
        line += [val.ljust(size)]

        for col in cols[he + 1:]:
            val = '' if row[col] is None else str(row[col])
            val = val.replace('|', '\\|')
            size = sizes[col]
            line.append(val.ljust(size))

        lines.append(line)

    lines = [' | '.join(line) for line in lines]
    lines = [l.rstrip() for l in lines]
    return '\n'.join(lines)


SHORT_NAMES = {
    'd': 'dataset',
    'r': 'resource',
    'b': 'base',
    'm': 'model',
    'p': 'property',
    't': 'type',
}


def normalizes_columns(
    cols: List[str],
    *,
    check_column_names: bool = True,
) -> List[ManifestColumn]:
    result: List[ManifestColumn] = []
    unknown: List[str] = []
    invalid: List[str] = []
    for col in cols:
        col = col or ''
        col = col.strip().lower()
        col = SHORT_NAMES.get(col, col)
        col = cast(ManifestColumn, col)
        if col not in MANIFEST_COLUMNS:
            unknown.append(col)
        else:
            if unknown:
                result += unknown
                invalid += unknown
                unknown = []
            result.append(col)
    if check_column_names and invalid:
        if len(invalid) == 1:
            raise PropertyNotFound(property=invalid[0])
        else:
            raise MultipleErrors(
                PropertyNotFound(property=col) for col in invalid
            )
    return result


def write_tabular_manifest(
    path: str,
    rows: Union[
        Manifest,
        Iterable[ManifestRow],
        None,
    ] = None,
    cols: List[ManifestColumn] = None,
) -> None:
    cols = cols or DATASET

    if rows is None:
        rows = []
    elif isinstance(rows, Manifest):
        rows = datasets_to_tabular(rows)

    rows = ({c: row[c] for c in cols} for row in rows)
    if path.endswith('.csv'):
        _write_csv(pathlib.Path(path), rows, cols)
    elif path.endswith('.xlsx'):
        _write_xlsx(pathlib.Path(path), rows, cols)
    else:
        raise ValueError(f"Unknown tabular manifest format {path!r}.")


def _write_csv(
    path: pathlib.Path,
    rows: Iterator[ManifestRow],
    cols: List[ManifestColumn],
) -> None:
    with path.open('w') as f:
        writer = csv.DictWriter(f, fieldnames=cols)
        writer.writeheader()
        writer.writerows(rows)


def _write_xlsx(
    path: pathlib.Path,
    rows: Iterator[ManifestRow],
    cols: List[ManifestColumn],
) -> None:
    workbook = xlsxwriter.Workbook(path, {
        'strings_to_formulas': False,
        'strings_to_urls': False,
    })

    bold = workbook.add_format({'bold': True})

    formats = {
        'id': workbook.add_format({
            'align': 'right',
            'valign': 'top',
        }),
        'dataset': workbook.add_format({
            'bold': True,
            'valign': 'top',
            'font_color': '#127622',
        }),
        'resource': workbook.add_format({
            'valign': 'top',
        }),
        'base': workbook.add_format({
            'valign': 'top',
        }),
        'model': workbook.add_format({
            'bold': True,
            'valign': 'top',
            'font_color': '#127622',
        }),
        'property': workbook.add_format({
            'valign': 'top',
            'font_color': '#127622',
        }),
        'type': workbook.add_format({
            'valign': 'top',
        }),
        'ref': workbook.add_format({
            'valign': 'top',
            'font_color': '#127622',
        }),
        'source': workbook.add_format({
            'valign': 'top',
            'font_color': '#c9211e',
        }),
        'prepare': workbook.add_format({
            'valign': 'top',
            'font_color': '#c9211e',
        }),
        'level': workbook.add_format({
            'valign': 'top',
        }),
        'access': workbook.add_format({
            'valign': 'top',
        }),
        'uri': workbook.add_format({
            'valign': 'top',
            'font_color': '#284f80',
        }),
        'title': workbook.add_format({
            'valign': 'top',
            'text_wrap': True,
        }),
        'description': workbook.add_format({
            'valign': 'top',
            'text_wrap': True,
        }),
    }

    sheet = workbook.add_worksheet()
    sheet.freeze_panes(1, 0)  # Freeze the first row.

    sheet.set_column('A:E', 2)   # id, d, r, b, m
    sheet.set_column('F:F', 20)  # property
    sheet.set_column('I:J', 20)  # source, prepare
    sheet.set_column('N:N', 20)  # title
    sheet.set_column('O:O', 30)  # description

    for j, col in enumerate(cols):
        sheet.write(0, j, col, bold)

    for i, row in enumerate(rows, 1):
        for j, col in enumerate(cols):
            val = row[col]
            fmt = formats.get(col)
            sheet.write(i, j, val, fmt)

    workbook.close()


def _get_state_obj(reader: TabularReader) -> Optional[TabularReader]:
    if reader is None or reader.name == "/":
        return None
    else:
        return reader



import re
from datetime import datetime
import pathlib
from typing import Any
from typing import Dict
from typing import List, Iterable, Optional
from typing import Type

import jsonpatch

from spinta import commands
from spinta.backends.components import BackendOrigin
from spinta.components import Mode
from spinta.backends.helpers import load_backend
from spinta.components import Model
from spinta.datasets.components import Dataset
from spinta.datasets.components import Entity
from spinta.datasets.components import Resource
from spinta.dimensions.enum.helpers import load_enums
from spinta.dimensions.prefix.helpers import load_prefixes
from spinta.exceptions import ManifestFileDoesNotExist
from spinta.exceptions import UnknownKeyMap
from spinta.manifests.components import ManifestSchema
from spinta.nodes import get_node
from spinta.core.config import RawConfig
from spinta.components import Context, Config, Store, MetaData, EntryId
from spinta.manifests.components import Manifest
from spinta.manifests.internal.components import InternalManifest
from spinta.types.namespace import load_namespace_from_name
from spinta.utils.enums import enum_by_name
from spinta.core.enums import Access
from spinta.utils.imports import importstr


def init_manifest(context: Context, manifest: Manifest, name: str):
    config = context.get('config')
    manifest.name = name
    manifest.store = context.get('store')
    manifest.parent = None
    manifest.endpoints = {}
    manifest.backends = {}
    manifest.objects = {name: {} for name in config.components['nodes']}
    manifest.sync = []
    manifest.prefixes = {}
    manifest.enums = {}
    manifest.access = Access.protected
    manifest.keymap = None
    manifest.backend = None
    manifest.mode = Mode.internal


def clone_manifest(context: Context, name: str = 'manifest') -> Manifest:
    manifest = Manifest()
    store: Store = context.get('store')
    init_manifest(context, manifest, name)
    if store.manifest:
        manifest.keymap = store.manifest.keymap
    return manifest


def create_manifest(
    context: Context,
    store: Store,
    name: str,
    seen: List[str] = None,
) -> Manifest:
    rc: RawConfig = context.get('rc')
    config = context.get('config')
    mtype = rc.get('manifests', name, 'type', required=True)
    Manifest_ = config.components['manifests'][mtype]
    manifest = Manifest_()
    manifest.type = mtype
    init_manifest(context, manifest, name)
    _configure_manifest(context, rc, store, manifest, seen)
    return manifest


def create_internal_manifest(context: Context, store: Store) -> InternalManifest:
    rc = context.get('rc')
    manifest = InternalManifest()
    manifest.type = 'yaml'
    init_manifest(context, manifest, 'internal')
    _configure_manifest(
        context, rc, store, manifest,
        backend=store.manifest.backend.name if store.manifest.backend else None,
    )
    return manifest


def _configure_manifest(
    context: Context,
    rc: RawConfig,
    store: Store,
    manifest: Manifest,
    seen: List[str] = None,
    *,
    backend: str = 'default',
):
    seen = seen or []
    name = manifest.name
    access = rc.get('manifests', name, 'access')
    if access:
        manifest.access = enum_by_name(manifest, 'access', Access, access)
    keymap = rc.get('manifests', name, 'keymap', default=None) or 'default'
    if keymap:
        if keymap not in store.keymaps:
            raise UnknownKeyMap(manifest, keymap=keymap)
        manifest.keymap = store.keymaps[keymap]
    backend = rc.get('manifests', name, 'backend', default=backend)
    if backend:
        manifest.backend = store.backends[backend]
    mode = rc.get('manifests', name, 'mode')
    if mode:
        manifest.mode = enum_by_name(manifest, 'mode', Mode, mode)
    for source in rc.get('manifests', name, 'sync', default=[], cast=list):
        if source in seen:
            raise Exception("Manifest sync cycle: " + ' -> '.join(seen + [source]))
        manifest.sync.append(
            create_manifest(context, store, source, seen + [source])
        )
    commands.configure(context, manifest)


def load_manifest_nodes(
    context: Context,
    manifest: Manifest,
    schemas: Iterable[ManifestSchema],
    *,
    source: Manifest = None,
    link: bool = False,
) -> None:
    to_link = []
    config = context.get('config')
    for eid, schema in schemas:
        if schema.get('type') == 'manifest':
            _load_manifest(context, manifest, schema, eid)
        else:
            node = _load_manifest_node(context, config, manifest, source, eid, schema)
            manifest.objects[node.type][node.name] = node
            if link:
                to_link.append(node)

    if '' not in manifest.namespaces:
        # Root namespace must always be present in manifest event if manifest is
        # empty.
        load_namespace_from_name(context, manifest, '', drop=False)

    if to_link:
        for node in to_link:
            commands.link(context, node)


def _load_manifest_backends(
    context: Context,
    manifest: Manifest,
    backends: Dict[str, Dict[str, str]],
) -> None:
    manifest.backends = {}
    for name, data in backends.items():
        manifest.backends[name] = load_backend(
            context,
            manifest,
            name,
            BackendOrigin.manifest,
            data,
        )


def _load_manifest(
    context: Context,
    manifest: Manifest,
    data: Dict[str, Any],
    eid: EntryId,
) -> None:
    if 'backends' in data:
        _load_manifest_backends(context, manifest, data['backends'])
    if 'prefixes' in data:
        prefixes = load_prefixes(context, manifest, manifest, data['prefixes'])
        manifest.prefixes.update(prefixes)
    if 'enums' in data:
        parents = [manifest]
        enums = load_enums(context, parents, data['enums'])
        manifest.enums.update(enums)


def _load_manifest_node(
    context: Context,
    config: Config,
    manifest: Manifest,
    source: Optional[Manifest],
    eid: EntryId,
    data: dict,
) -> MetaData:
    node = get_node(config, manifest, eid, data)
    node.eid = eid
    node.type = data['type']
    node.parent = manifest
    node.manifest = manifest
    commands.load(context, node, data, manifest, source=source)
    return node


def get_current_schema_changes(
    context: Context,
    manifest: Manifest,
    eid: EntryId,
) -> List[dict]:
    freezed = commands.manifest_read_freezed(context, manifest, eid=eid)
    current = commands.manifest_read_current(context, manifest, eid=eid)
    patch = jsonpatch.make_patch(freezed, current)
    return list(patch)


def entity_to_schema(entity: Optional[Entity]) -> ManifestSchema:
    if entity is None:
        return None, {
            'dataset': None,
            'resource': None,
            'name': '',
            'pk': [],
        }
    else:
        return None, {
            'dataset': entity.dataset.name if entity.dataset else None,
            'resource': entity.resource.name if entity.resource else None,
            'name': entity.name,
            'prepare': entity.prepare,
            'pk': [p.name for p in entity.pkeys],
            # 'pk': [
            #     to_property_name(p)
            #     for p in _ensure_list(schema.primary_key)
            # ],
        }


def model_to_schema(model: Optional[Model]) -> ManifestSchema:
    if model is None:
        return None, {
            'type': 'model',
            'name': '',
            'level': None,
            'access': None,
            'title': '',
            'description': '',
            'external': None,
            'properties': {},
        }
    else:
        return model.eid, {
            'type': 'model',
            'name': model.name,
            'level': model.level.name if model.level else None,
            'access': model.given.access,
            'title': model.title,
            'description': model.description,
        }


def resource_to_schema(resource: Resource) -> ManifestSchema:
    if (
        resource.backend and
        resource.backend.origin != BackendOrigin.resource
    ):
        backend = resource.backend.name
    else:
        backend = ''

    return resource.eid, {
        'type': resource.type,
        'backend': backend,
        'external': resource.external,
        'prepare': None if resource.prepare is None else str(resource.prepare),
        'level': resource.level.name if resource.level else None,
        'access': resource.given.access,
        'title': resource.title,
        'description': resource.description,
    }


def dataset_to_schema(dataset: Dataset) -> ManifestSchema:
    return dataset.eid, {
        'type': 'dataset',
        'name': dataset.name,
        'level': dataset.level.name if dataset.level else None,
        'access': dataset.given.access,
        'title': dataset.title,
        'description': dataset.description,
        'resources': {
            resource.name: {
                'type': resource.type,
                'backend': (
                    resource.backend.name
                    if (
                        resource.backend and
                        resource.backend.origin != BackendOrigin.resource
                    )
                    else ''
                ),
                'level': resource.level.name if resource.level else None,
                'access': resource.given.access,
                'title': resource.title,
                'description': resource.description,
            }
            for resource in dataset.resources.values()
        }
    }


def detect_manifest_from_path(rc: RawConfig, path: str) -> Type[Manifest]:
    for name in rc.keys('components', 'manifests'):
        Manifest_: Type[Manifest] = rc.get(
            'components', 'manifests', name,
            cast=importstr,
        )
        if Manifest_.detect_from_path(path):
            return Manifest_
    raise ValueError(
        f"Can't find manifest type matching given path {path!r}"
    )


def get_manifest_from_type(rc: RawConfig, type_: str) -> Type[Manifest]:
    for name in rc.keys('components', 'manifests'):
        Manifest_: Type[Manifest] = rc.get(
            'components', 'manifests', name,
            cast=importstr,
        )
        if Manifest_.type == type_:
            return Manifest_
    raise ValueError(
        f"Can't find manifest component matching given type {type_!r}"
    )


def check_manifest_path(manifest: Manifest, path: str) -> None:
    if (
        not path.startswith(('http://', 'https://')) and
        not pathlib.Path(path).exists()
    ):
        raise ManifestFileDoesNotExist(manifest, path=path)


class TypeDetector:
    _binary_pattern = re.compile('[01]+')

    def __init__(self):
        self._values = set()
        self.type = ""
        self.unique = True
        self.required = True
        self.array = False

    def get_type(self) -> str:
        value = self.type
        if self.type == "":
            value = "string"
        return value

    def detect(self, value: Any) -> None:
        self._process_for_unique(value)
        self._process_for_required(value)

        if value is not None and value != "":
            if not isinstance(value, (list, tuple)):
                value = [value]
            for item in value:
                str_value = str(item)
                if self.type in self._key_list:
                    self._type_detector_priority[self.type](self, str_value)

                if self.type == "":
                    for key in self._key_list[self._highest_type_level:]:
                        self._type_detector_priority[key](self, str_value)
                        if self.type != "":
                            current_level_index = self._key_list.index(self.type)
                            if current_level_index > self._highest_type_level:
                                self._highest_type_level = current_level_index
                            break

    def _process_for_unique(self, value: Any):
        if self.unique:
            if value in self._values:
                self.unique = False
            else:
                self._values.add(value)

    def _process_for_required(self, value: Any):
        if self.required:
            if value is None:
                self.required = False
            elif value == '':
                self.required = False

    def _assert_boolean(self, value: str):
        if value in ["0", "1", "true", "false"]:
            self.type = "boolean"
        else:
            self.type = ""

    def _assert_number(self, value: str):
        if value.strip("-").replace(".", "", 1).isnumeric():
            self.type = "number"
        else:
            self.type = ""

    def _assert_integer(self, value: str):
        if value.strip("-").isnumeric():
            self.type = "integer"
        else:
            self.type = ""

    def _assert_binary(self, value: str):
        match = self._binary_pattern.fullmatch(value)
        if match:
            self.type = "binary"
        else:
            self.type = ""

    def _assert_date(self, value: str):
        try:
            datetime.strptime(value, '%Y-%m-%d')
            self.type = "date"
        except:
            self.type = ""

    def _assert_datetime(self, value: str):
        try:
            dt_str = value
            datetime.fromisoformat(dt_str.replace('Z', '+00:00'))
            self.type = "datetime"
        except:
            self.type = ""

    def _assert_time(self, value: str):
        formats = [
            '%H:%M:%S.%f%z',
            '%H:%M:%S.%f',
            '%H:%M:%S%z',
            '%H:%M:%S',
            '%H:%M%z',
            '%H:%M'
        ]
        new_type = ""
        for time_format in formats:
            try:
                datetime.strptime(value, time_format)
                new_type = "time"
                break
            except:
                continue
        self.type = new_type

    def _assert_url(self, value: str):
        if value.startswith(("http://", "https://")):
            self.type = "url"
        else:
            self.type = ""

    def _assert_string(self, value: str):
        self.type = "string"

    _type_detector_priority = {
        "boolean": _assert_boolean,
        "binary": _assert_binary,
        "integer": _assert_integer,
        "number": _assert_number,
        "time": _assert_time,
        "date": _assert_date,
        "datetime": _assert_datetime,
        "url": _assert_url,
        "string": _assert_string
    }
    _key_list = list(_type_detector_priority.keys())
    _highest_type_level = 0



from typing import overload
import sqlalchemy as sa

from sqlalchemy.dialects.postgresql import JSONB, UUID

from spinta import commands
from spinta.components import Context, Model
from spinta.core.enums import Access
from spinta.manifests.components import Manifest
from spinta.types.datatype import DataType, PrimaryKey, Ref, BackRef
from spinta.backends.constants import TableType
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.constants import UNSUPPORTED_TYPES
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name
from spinta.backends.postgresql.helpers import get_column_name
from spinta.backends.postgresql.helpers.changes import get_changes_table
from spinta.datasets.enums import Level


@overload
@commands.prepare.register(Context, PostgreSQL, Manifest)
def prepare(context: Context, backend: PostgreSQL, manifest: Manifest):
    # Prepare backend for models.
    for model in manifest.models.values():
        if model.backend and model.backend.name == backend.name:
            commands.prepare(context, backend, model)


@overload
@commands.prepare.register(Context, PostgreSQL, Model)
def prepare(context: Context, backend: PostgreSQL, model: Model):
    columns = []
    for prop in model.properties.values():
        # FIXME: _revision should has its own type and on database column type
        #        should bet received from get_primary_key_type() command.
        if prop.name.startswith('_') and prop.name not in ('_id', '_revision'):
            continue

        column = commands.prepare(context, backend, prop)

        if isinstance(column, list):
            columns.extend(column)
        elif column is not None:
            columns.append(column)

    if model.unique:
        for constraint in model.unique:
            prop_list = []
            for prop in constraint:
                name = prop.name
                if isinstance(prop.dtype, Ref):
                    if prop.level is None or prop.level > Level.open:
                        name = f'{name}._id'
                    elif prop.dtype:
                        name = f'{name}.{prop.dtype.refprops[0].name}'
                prop_list.append(name)

            columns.append(sa.UniqueConstraint(*prop_list))

    # Create main table.
    main_table_name = get_pg_name(get_table_name(model))
    pkey_type = commands.get_primary_key_type(context, backend)
    main_table = sa.Table(
        main_table_name, backend.schema,
        sa.Column('_txn', pkey_type, index=True),
        sa.Column('_created', sa.DateTime),
        sa.Column('_updated', sa.DateTime),
        *columns,
    )
    backend.add_table(main_table, model)
    # Create changes table.
    changelog_table = get_changes_table(context, backend, model)
    backend.add_table(changelog_table, model, TableType.CHANGELOG)


@overload
@commands.prepare.register(Context, PostgreSQL, DataType)
def prepare(context: Context, backend: PostgreSQL, dtype: DataType):
    if dtype.name in UNSUPPORTED_TYPES:
        return
    prop = dtype.prop
    name = get_column_name(prop)
    types = {
        'string': sa.Text,
        'text': JSONB,
        'date': sa.Date,
        'time': sa.Time,
        'datetime': sa.DateTime,
        'temporal': sa.DateTime,
        'integer': sa.Integer,
        'number': sa.Float,
        'boolean': sa.Boolean,
        'binary': sa.LargeBinary,
        'json': JSONB,
        'spatial': sa.Text,  # unsupported
        'image': sa.Text,  # unsupported
        'url': sa.String,
        'uri': sa.String,
        'denorm': sa.String,
    }

    if dtype.name not in types:
        raise Exception(
            f"Unknown type {dtype.name!r} for property {prop.place!r}."
        )
    column_type = types[dtype.name]
    nullable = not dtype.required
    return sa.Column(name, column_type, unique=dtype.unique, nullable=nullable)


@commands.get_primary_key_type.register()
def get_primary_key_type(context: Context, backend: PostgreSQL):
    return UUID()


@overload
@commands.prepare.register(Context, PostgreSQL, PrimaryKey)
def prepare(context: Context, backend: PostgreSQL, dtype: PrimaryKey):
    pkey_type = commands.get_primary_key_type(context, backend)
    base = dtype.prop.model.base
    if base and (base.level and base.level >= Level.identifiable or not base.level):
        return [
            sa.Column('_id', pkey_type, primary_key=True),
            sa.ForeignKeyConstraint(
                ['_id'], [f'{get_pg_name(get_table_name(base.parent))}._id'],
                name=get_pg_name(f'fk_{base.parent.name}_id'),
            )
        ]
    return sa.Column('_id', pkey_type, primary_key=True)



class ErrorCounter:
    count: int
    max_error_count: int

    def __init__(self, max_count: int):
        self.max_error_count = max_count
        self.reset()

    def increase(self):
        self.count += 1

    def reset(self):
        self.count = 0

    def has_errors(self) -> bool:
        return self.count > 0

    def has_reached_max(self) -> bool:
        return self.count >= self.max_error_count




from spinta.core.ufuncs import ufunc
from spinta.dimensions.enum.components import EnumFormula
from spinta.exceptions import FormulaError


@ufunc.resolver(EnumFormula, str)
def bind(env: EnumFormula, name: str) -> None:
    raise FormulaError(env.node, formula=name, error="Binds are not supported.")




from typing import Dict, Any

import pymongo

from spinta import commands
from spinta.components import Context
from spinta.backends.mongo.components import Mongo


@commands.load.register(Context, Mongo, dict)
def load(context: Context, backend: Mongo, config: Dict[str, Any]):
    # Load Mongo client using configuration.
    backend.dsn = config['dsn']
    backend.db_name = config['db']
    backend.client = pymongo.MongoClient(backend.dsn)
    backend.db = backend.client[backend.db_name]



import collections
import datetime
import itertools
from typing import AsyncIterator
from typing import Iterable
from typing import Iterator
from typing import Tuple
from typing import cast

from toposort import toposort

from spinta import commands
from spinta.commands.write import push_stream
from spinta.commands.write import write
from spinta.components import Action
from spinta.components import Context
from spinta.components import DataItem
from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import bind
from spinta.manifests.backend.components import BackendManifest
from spinta.manifests.components import Manifest
from spinta.utils.aiotools import adrain
from spinta.utils.aiotools import aiter
from spinta.utils.itertools import last
from spinta.utils.json import fix_data_for_json


async def run_bootstrap(context: Context, manifest: BackendManifest):
    # Bootstrap all backends
    for backend in manifest.store.backends.values():
        commands.bootstrap(context, backend)

    # Sync versions
    stream = read_sync_versions(context, manifest)
    stream = versions_to_dstream(manifest, stream, applied=True)
    await adrain(push_stream(context, stream))

    # Update schemas to last version
    for schema in read_lastest_version_schemas(context, manifest):
        await update_schema_version(context, manifest, schema)


async def run_migrations(context: Context, manifest: BackendManifest):
    # Sync versions
    stream = read_sync_versions(context, manifest)
    stream = versions_to_dstream(manifest, stream)
    await adrain(push_stream(context, stream))

    # Apply unapplied versions
    store = manifest.store
    model = manifest.objects['ns']['']
    backends = {}
    versions = read_unapplied_versions(context, manifest)
    versions = itertools.groupby(versions, key=lambda v: v.get('backend', 'default'))
    for name, group in versions:
        if name not in backends:
            backend = store.backends[name]
            context.attach(f'transaction.{backend.name}', backend.begin)
            backends[backend] = commands.migrate(context, manifest, backend)
        else:
            backend = None
        execute = backends[backend]
        async for version in execute(group):
            schema = version['schema']
            assert schema['version'] == version['id'], (
                schema['version'],
                version['id'],
            )
            await write(context, model, [
                {
                    '_op': 'patch',
                    '_type': '_schema/version',
                    '_where': '_id="%s"' % version['id'],
                    'applied': datetime.datetime.now(datetime.timezone.utc),
                },
                {
                    '_op': 'upsert',
                    '_type': '_schema',
                    '_where': '_id="%s"' % schema['id'],
                    '_id': schema['id'],
                    'type': schema['type'],
                    'name': schema['name'],
                    'version': schema['version'],
                    'schema': fix_data_for_json(schema),
                },
            ])


def read_unapplied_versions(
    context: Context,
    manifest: Manifest,
):
    model = manifest.objects['model']['_schema/version']
    query = Expr(
        'and',
        Expr('select', bind('id'), bind('_id'), bind('parents')),
        Expr('ne', bind('applied'), None),
    )
    versions = {}
    for row in commands.getall(context, model, model.backend, query=query):
        versions[row['_id']] = set(row['parents'])
    for group in toposort(versions):
        for vid in sorted(group):
            yield get_version_schema(context, manifest, vid)


def read_sync_versions(context: Context, manifest: Manifest):
    sources = [cast(Manifest, manifest.store.internal)] + manifest.sync
    for source in sources:
        for eid in commands.manifest_list_schemas(context, source):
            for version in commands.manifest_read_versions(context, source, eid=eid):
                yield version


async def versions_to_dstream(
    manifest: BackendManifest,
    versions: Iterable[dict],
    *,
    applied: bool = False,
) -> AsyncIterator[DataItem]:
    now = datetime.datetime.now(datetime.timezone.utc)
    model = manifest.objects['model']['_schema/version']
    for version in versions:
        payload = {
            '_op': 'upsert',
            '_where': '_id="%s"' % version['id'],
            '_id': version['id'],
            'type': version['schema']['type'],
            'name': version['schema']['name'],
            'id': version['schema']['id'],
            'created': version['date'],
            'synced': now,
            'applied': None,
            'parents': version['parents'],
            'schema': fix_data_for_json(version['schema']),
            'migrate': version['migrate'],
            'changes': fix_data_for_json(version['changes']),
        }
        if applied:
            payload['applied'] = now
        yield DataItem(model, action=Action.UPSERT, payload=payload)


def list_schemas(context: Context, manifest: BackendManifest):
    model = manifest.objects['model']['_schema']
    query = {
        'select': ['_id'],
    }
    for row in commands.getall(context, model, model.backend, **query):
        yield row['_id']


def read_schema(context: Context, manifest: BackendManifest, eid: str):
    model = manifest.objects['model']['_schema']
    row = commands.getone(context, model, model.backend, id_=eid)
    return row['schema']


def list_sorted_unapplied_versions(
    context: Context,
    manifest: Manifest,
) -> Iterator[Tuple[str, str]]:
    model = manifest.objects['model']['_schema/version']
    query = {
        'select': ['id', '_id', 'parents'],
        'query': [
            {'name': 'eq', 'args': ['applied', None]}
        ],
    }
    schemas = {}
    versions = {}
    for row in commands.getall(context, model, model.backend, **query):
        schemas[row['_id']] = row['id']
        versions[row['_id']] = row['parents']

    for group in toposort(versions):
        for vid in sorted(group):
            yield schemas[vid], vid


def read_lastest_version_schemas(
    context: Context,
    manifest: Manifest,
) -> Iterator[Tuple[str, str]]:
    model = manifest.objects['model']['_schema/version']
    query = Expr(
        'and',
        Expr('select', bind('id'), bind('_id'), bind('parents')),
        Expr('ne', bind('applied'), None),
    )
    schemas = collections.defaultdict(dict)
    for row in commands.getall(context, model, model.backend, query=query):
        schemas[row['id']][row['_id']] = set(row['parents'])

    for schema_id, versions in schemas.items():
        last_version = last(toposort(versions))
        assert len(last_version) == 1, last_version
        last_version = next(iter(last_version))
        yield get_version_schema(context, manifest, last_version)


def get_last_version_eid(
    context: Context,
    manifest: Manifest,
    schema_eid: str,
) -> Iterator[Tuple[str, str]]:
    model = manifest.objects['model']['_schema/version']
    query = Expr(
        'and',
        Expr('select', bind('_id'), bind('parents')),
        Expr('eq', bind('id'), schema_eid),
        Expr('ne', bind('applied'), None),
    )
    versions = {}
    for row in commands.getall(context, model, model.backend, query=query):
        versions[row['_id']] = row['parents']
    last_version = last(toposort(versions))
    assert len(last_version) == 1, last_version
    return last_version[0]


def get_version_schema(
    context: Context,
    manifest: Manifest,
    version_eid: str,
) -> Iterator[Tuple[str, str]]:
    model = manifest.objects['model']['_schema/version']
    version = commands.getone(context, model, model.backend, id_=version_eid)
    return version['schema']


async def update_schema_version(context: Context, manifest: Manifest, schema: dict):
    model = manifest.objects['model']['_schema']
    data = DataItem(model, action=Action.UPSERT, payload={
        '_op': 'upsert',
        '_where': '_id="%s"' % schema['id'],
        '_id': schema['id'],
        'type': schema['type'],
        'name': schema['name'],
        'version': {
            '_id': schema['version']
        },
        'schema': fix_data_for_json(schema),
    })
    await adrain(push_stream(context, aiter([data])))



import contextlib

from spinta.datasets.components import ExternalBackend


class DaskBackend(ExternalBackend):
    type: str = 'dask'

    @contextlib.contextmanager
    def begin(self):
        yield


class Xml(DaskBackend):
    type: str = 'xml'

    @contextlib.contextmanager
    def begin(self):
        yield

        
class Csv(DaskBackend):
    type: str = 'csv'

    @contextlib.contextmanager
    def begin(self):
        yield


class Json(DaskBackend):
    type: str = 'json'

    @contextlib.contextmanager
    def begin(self):
        yield




from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest


@commands.check.register(Context, Manifest)
def check(context: Context, manifest: Manifest):
    for objects in manifest.objects.values():
        for obj in objects.values():
            check(context, obj)

    # Check endpoints.
    names = set(manifest.models)
    for model in manifest.models.values():
        if model.endpoint == model.name or model.endpoint in names:
            raise Exception(
                f"Endpoint name can't overshadow existing model names and "
                f"{model.endpoint!r} is already a model name."
            )



from __future__ import annotations

from typing import Dict
from typing import List
from typing import Optional

import sqlalchemy as sa
from sqlalchemy.engine.base import Engine

from spinta.backends.components import Backend
from spinta.components import EntryId
from spinta.components import Namespace
from spinta.dimensions.lang.components import LangData
from spinta.components import MetaData
from spinta.components import Model
from spinta.components import Node
from spinta.components import Property
from spinta.core.enums import Access
from spinta.core.ufuncs import Expr
from spinta.datasets.enums import Level
from spinta.dimensions.prefix.components import UriPrefix
from spinta.manifests.components import Manifest
from spinta.types.owner import Owner
from spinta.types.project import Project


class DatasetGiven:
    access: str = None


class Dataset(MetaData):
    """DCAT compatible metadata about an external dataset.

    DCAT (Data Catalog Vocabulary) - https://w3c.github.io/dxwg/dcat/
    """

    id: str
    manifest: Manifest
    owner: Owner = None
    level: Level = 3
    access: Access = Access.private
    website: str = None
    projects: List[Project] = None
    resources: Dict[str, Resource] = None
    source: Optional[str] = None  # metadata source
    title: str
    description: str
    given: DatasetGiven
    lang: LangData = None
    ns: Namespace = None
    prefixes: Dict[str, UriPrefix]

    schema = {
        'type': {'type': 'string', 'required': True},
        'name': {'type': 'string', 'required': True},
        'path': {'type': 'string'},
        'version': {
            'type': 'object',
            'items': {
                'number': {'type': 'integer', 'required': True},
                'date': {'type': 'date', 'required': True},
            },
        },

        'manifest': {'parent': True},
        'owner': {'type': 'ref', 'ref': 'manifest.nodes.owner'},
        'level': {'type': 'integer', 'choices': Level},
        'access': {'type': 'string', 'choices': Access},
        'website': {'type': 'url'},
        'projects': {
            'type': 'array',
            'factory': list,
            'items': {
                'type': 'ref',
                'ref': 'manifest.nodes.project',
            },
        },
        'resources': {
            'type': 'object',
            'values': {
                'type': 'comp',
                'group': 'external',
                'ctype': 'resource',
            }
        },
        'source': {'type': 'string'},
    }

    def __init__(self):
        self.given = DatasetGiven()
        self.prefixes = {}


class ExternalBackend(Backend):
    engine: Engine = None
    schema: sa.MetaData = None


class External(Node):
    pass


class ResourceGiven:
    access: Access = None


class Resource(External):
    eid: EntryId = None
    title: str
    description: str
    dataset: Dataset
    backend: ExternalBackend = None
    level: Level
    access: Access
    external: str
    prepare: str
    models: Dict[str, Model]
    given: ResourceGiven
    lang: LangData = None
    comments: List[Comment] = None

    schema = {
        'type': {'type': 'string'},
        'dataset': {'parent': True},
        'prepare': {'type': 'spyna'},

        # Backend name specified in `ref` column, points to previously defined
        # backend or to a configured stored backend.
        'backend': {
            'type': 'ref',
            'ref': 'context.store.backends',
        },

        # If `backend` is not given via `ref` column, then in `source` column,
        # explicit backend dsn can be given. Only one `ref` or `source` can be
        # given.
        'external': {
            'type': 'string'
        },
        'params': {'type': 'object'},
        'level': {
            'type': 'integer',
            'choices': Level,
            'inherit': 'dataset.level',
        },
        'access': {
            'type': 'string',
            'choices': Access,
            'inherit': 'dataset.access',
        },
        'title': {'type': 'string'},
        'description': {'type': 'string'},
        'comments': {},
    }

    def __init__(self):
        self.given = ResourceGiven()
        self.params = {}


class Param:
    name: str
    sources: List[str]
    formulas: List[Expr]


class Entity(External):
    # XXX: Here `dataset` and `resource` first are resolved as `str` and then
    #      linked with `Dataset` and `Resource` instances. That is not a good
    #      thing and should be refactored into a `RawEntity` with `str` types
    #      which then is converted into `Entity` with `Dataset` and `Resource`
    #      types.
    dataset: Dataset                # dataset
    resource: Optional[Resource]    # resource
    model: Model                    # model
    pkeys: List[Property]           # model.ref
    # This is set to True if primary key is not given. Anf if primary key is not
    # given, then `pkeys` will be set to all properties.
    unknown_primary_key: bool = False
    name: str                       # model.source
    prepare: Expr                   # model.prepare
    params: List[Param]

    schema = {
        'model': {'parent': True},
        'dataset': {'type': 'ref', 'ref': 'context.nodes.dataset'},
        'resource': {'type': 'ref', 'ref': 'dataset.resources'},
        'name': {'type': 'string', 'default': None},
        'prepare': {'type': 'spyna', 'default': None},
        'params': {
            'type': 'array',
            'items': {'type': 'object'},
        },
        'pk': {
            'type': 'array',
            'force': True,
            'attr': 'pkeys',
            'items': {
                'type': 'ref',
                'ref': 'model.properties',
            },
        }
    }


class Attribute(External):
    prop: Property          # property
    name: str               # property.source
    prepare: Expr = None    # property.prepare

    schema = {
        'prop': {'parent': True},
        'name': {'default': None},
        'prepare': {'type': 'spyna', 'default': None},
    }



import hashlib

from spinta.components import Property
from spinta.backends.postgresql.constants import NAMEDATALEN


def get_column_name(prop: Property) -> str:
    if prop.list:
        if prop.place == prop.list.place:
            return prop.list.name
        else:
            return prop.place[len(prop.list.place) + 1:]
    else:
        return prop.place


def get_pg_name(name: str) -> str:
    if len(name) > NAMEDATALEN:
        hs = 8
        h = hashlib.sha1(name.encode()).hexdigest()[:hs]
        i = int(NAMEDATALEN / 100 * 60)
        j = NAMEDATALEN - i - hs - 2
        name = name[:i] + '_' + h + '_' + name[-j:]
    return name


def get_pg_sequence_name(name: str) -> str:
    suffix = '__id_seq'
    slen = len(suffix)
    nlen = len(name)
    if nlen + slen > NAMEDATALEN:
        trim = (nlen + slen) - NAMEDATALEN
        name = name[:-trim]
    return name + suffix



from typing import List, Optional, Tuple


def get_widths(
    rows,
    cols: List[str],
    max_value_length: int = 100,
    max_col_width: Optional[int] = None,
    rows_to_check: int = 200
) -> Tuple[List[dict], dict]:
    widths = {}
    max_widths = {}
    max_split_widths = {}
    read_rows = []

    for i, row in enumerate(rows):
        read_rows.append(row)
        if i >= rows_to_check:
            break
        else:
            for col in cols:
                value = row.get(col)
                if value:
                    value = str(value)
                    if len(value) > max_value_length:
                        value = value[:max_value_length] + "..."

                    if len(value.splitlines()) > 1:
                        value_width = max([len(part) for part in value.splitlines()])
                    else:
                        value_width = len(value)

                    width = max(value_width, len(col))
                    split_width = max(len(value.split(' ')[0]), len(col))
                else:
                    width = len(col)
                    split_width = len(col)

                if col not in max_widths or width > max_widths[col]:
                    max_widths[col] = width
                if col not in max_split_widths or split_width > max_split_widths[col]:
                    max_split_widths[col] = split_width

                if col not in widths:
                    widths[col] = sum([max_widths[col], max_split_widths[col]]) // 2
                else:
                    widths[col] = sum([widths[col], max_widths[col], max_split_widths[col]]) // 3

                if max_col_width and widths[col] > max_col_width:
                    widths[col] = max_col_width

    return read_rows, widths


def get_displayed_cols(
    widths: dict,
    max_width: int,
    separator: str = "  "
) -> Tuple[bool, List[str]]:
    total_width = 0
    cols = []
    shortened = False
    for col, width in widths.items():
        total_width += width + len(separator)
        if total_width > max_width:
            shortened = True
            break
        else:
            cols.append(col)
    return shortened, cols


def draw_border(
    widths: dict,
    displayed_cols: List[str],
    separator: str = "  ",
    shortened: bool = False,
) -> str:
    borders = []
    for col in displayed_cols:
        width = widths.get(col)
        border = '-' * width
        borders.append(border)
    result = separator.join(borders)
    if shortened:
        result += separator + "..."
    return result + "\n"


def draw_header(
    widths: dict,
    displayed_cols: List[str],
    separator: str = "  ",
    shortened: bool = False,
) -> str:
    headers = []
    for col in displayed_cols:
        width = widths.get(col)
        if width > len(col):
            offset = width - len(col)
            col += ' ' * offset
        headers.append(col)
    result = separator.join(headers)
    if shortened:
        result += separator + "..."
    return result + "\n"


def draw_row(
    data: dict,
    widths: dict,
    displayed_cols: List[str],
    max_value_length: int = 100,
    separator: str = "  ",
    shortened: bool = False
) -> str:
    lines = {}
    line_count = 1
    lines[line_count] = prepare_line(widths, displayed_cols)

    for col in displayed_cols:
        width = widths.get(col)
        value = data.get(col)
        value = str(value) if value is not None else "∅"
        line_count = 1

        if len(value) > max_value_length:
            value = value[:max_value_length] + "..."

        if len(value.splitlines()) > 1:
            split_value = value.splitlines()
            for i, part in enumerate(split_value):
                part = part.strip()
                lines, line_count = get_row_lines(
                    col,
                    part,
                    width,
                    lines,
                    line_count,
                    widths,
                    displayed_cols
                )

                if i < (len(split_value) - 1):
                    line_count += 1
                    if line_count not in lines:
                        lines[line_count] = prepare_line(widths, displayed_cols)
        else:
            lines, _ = get_row_lines(
                col,
                value,
                width,
                lines,
                line_count,
                widths,
                displayed_cols
            )

    result = ""
    for i, row in enumerate(lines.values()):
        ending = ""
        if shortened:
            ending += separator + "..."
        if i < (len(lines.values()) - 1):
            ending += "\\"

        if not ending:
            row = separator.join(row.values()).rstrip()
        else:
            row = separator.join(row.values())
        ending += '\n'
        result += row + ending

    return result


def prepare_line(
    widths: dict,
    displayed_cols: List[str]
) -> dict:
    return {col: ' ' * widths[col] for col in widths.keys() if col in displayed_cols}


def get_row_lines(
    col: str,
    value: str,
    width: int,
    lines: dict,
    line_count: int,
    widths: dict,
    displayed_cols: List[str]
) -> Tuple[dict, int]:
    if len(value) > width:
        split_value = [value[i:i + width] for i in range(0, len(value), width)]
        for i, part in enumerate(split_value):
            part = part.strip()
            if width > len(part):
                offset = width - len(part)
                part += ' ' * offset
            lines[line_count][col] = part

            if i < (len(split_value) - 1):
                line_count += 1
                if line_count not in lines:
                    lines[line_count] = prepare_line(widths, displayed_cols)
    else:
        if width > len(value):
            offset = width - len(value)
            value += ' ' * offset

        lines[line_count][col] = value

    return lines, line_count



import datetime
from typer import echo
from typing import List

from spinta.auth import authorized
from spinta.cli.helpers.errors import ErrorCounter
from spinta.components import Context, Model, Action, Property
from spinta.datasets.backends.helpers import extract_values_from_row
from spinta.datasets.keymaps.components import KeyMap
import tqdm

from spinta.exceptions import UnauthorizedKeymapSync
from spinta.utils.response import get_request


def sync_keymap(context: Context, keymap: KeyMap, client, server: str, models: List[Model], error_counter: ErrorCounter, no_progress_bar: bool, reset_cid: bool):
    counters = {}
    if not no_progress_bar:
        counters = {
            '_total': tqdm.tqdm(desc='SYNCHRONIZING KEYMAP', ascii=True)
        }
        echo()

    for model in models:
        model_keymaps = [model.model_type()]
        primary_keys = model.external.pkeys
        skip_model = False
        for key in primary_keys:
            is_authorized = authorized(context, key, action=Action.SEARCH)
            if not is_authorized:
                echo(f"SKIPPED '{model.model_type()}' MODEL SYNC, NO PERMISSION.")
                skip_model = True
                break
        if skip_model:
            continue

        for combination in model.required_keymap_properties:
            model_keymaps.append(f'{model.model_type()}.{"_".join(combination)}')

        # Get min cid (in case new model was added, or models are out of sync)
        sync_cid = 0
        initial = True
        if not reset_cid:
            for keymap_name in model_keymaps:
                cid = keymap.get_sync_data(keymap_name)
                if not cid:
                    sync_cid = 0
                    break
                else:
                    if initial:
                        sync_cid = cid
                        initial = False
                    else:
                        sync_cid = min(sync_cid, cid)

        url = f'{server}/{model.model_type()}/:changes'
        if sync_cid:
            url = f'{url}/{sync_cid+1}'
        status_code, resp = get_request(
            client,
            url,
            ignore_errors=[404],
            error_counter=error_counter,
        )
        if status_code == 200:
            data = resp['_data']

            if not no_progress_bar:
                model_counters = {}
                for keymap_name in model_keymaps:
                    model_counters[keymap_name] = tqdm.tqdm(desc=keymap_name, ascii=True)
                counters[model.model_type()] = model_counters

            for row in data:
                if row['_op'] == 'insert':
                    sync_model_insert(keymap, model, row, primary_keys, counters)
                elif row['_op'] == 'patch' or row['_op'] == 'update':
                    sync_model_update(keymap, model, row, counters)
                sync_cid = row['_cid']

            for keymap_name in model_keymaps:
                keymap.update_sync_data(keymap_name, sync_cid, datetime.datetime.now())

            if model.model_type() in counters:
                for counter in counters[model.model_type()].values():
                    counter.close()

    if '_total' in counters:
        counters['_total'].close()


def remap_decoded_values(decoded: object, properties: List[str]):
    mapped = {}
    if not isinstance(decoded, list):
        decoded = [decoded]
    for i, prop in enumerate(properties):
        mapped[prop] = decoded[i]
    return mapped


def keymaps_affected_by_change(row: dict, model: Model) -> dict:
    keymaps = {}
    for key in model.external.pkeys:
        if key.name in row.keys():
            keymaps[model.model_type()] = list([prop.name for prop in model.external.pkeys])
            break

    for combination in model.required_keymap_properties:
        for prop in combination:
            if prop in row.keys():
                keymaps[f'{model.model_type()}.{"_".join(combination)}'] = combination
                break

    return keymaps


def _extract_row_data_from_keys(row: dict, keys: List[Property]) -> dict:
    result = {}
    for key in keys:
        if key.name in row:
            result[key.name] = row[key.name]
    return result


def sync_model_insert(keymap: KeyMap, model: Model, row: dict, primary_keys: List[Property], counters: dict):
    values = _extract_row_data_from_keys(row, primary_keys)
    value = list(values.values())
    if len(value) == 1:
        value = value[0]

    keymap.synchronize(model.model_type(), value, row['_id'])
    if model.model_type() in counters and model.model_type() in counters[model.model_type()]:
        counters[model.model_type()][model.model_type()].update(1)
    if '_total' in counters:
        counters['_total'].update(1)

    if row['_id'] and model.required_keymap_properties:
        for combination in model.required_keymap_properties:
            joined = '_'.join(combination)
            key = f'{model.model_type()}.{joined}'
            val = extract_values_from_row(row, model, combination)
            keymap.synchronize(key, val, row['_id'])
            if model.model_type() in counters and key in counters[model.model_type()]:
                counters[model.model_type()][key].update(1)
            if '_total' in counters:
                counters['_total'].update(1)


def sync_model_update(keymap: KeyMap, model: Model, row: dict, counters: dict):
    for km, props in keymaps_affected_by_change(row, model).items():
        decoded = keymap.decode(km, row['_id'])
        remapped = remap_decoded_values(decoded, props)
        for key in remapped.keys():
            if key in row:
                remapped[key] = row[key]
        val = list(remapped.values())
        if len(val) == 1:
            val = val[0]
        keymap.synchronize(km, val, row['_id'])
        if model.model_type() in counters and km in counters[model.model_type()]:
            counters[model.model_type()][km].update(1)
        if '_total' in counters:
            counters['_total'].update(1)



from typing import Optional, Any

import datetime
import uuid
import hashlib
import decimal
from uuid import UUID

import msgpack
import sqlalchemy as sa
from sqlalchemy.dialects.sqlite import insert

from spinta import commands
from spinta.cli.helpers.data import ensure_data_dir
from spinta.components import Config
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.datasets.keymaps.components import KeyMap
from sqlalchemy.dialects.sqlite import insert


class SqlAlchemyKeyMap(KeyMap):
    dsn: str = None

    def __init__(self, dsn: str = None):
        self.dsn = dsn
        self.engine = None
        self.metadata = None
        self.conn = None

    def __enter__(self):
        assert self.dsn is not None
        assert self.conn is None
        self.conn = self.engine.connect()
        return self

    def __exit__(self, *exc):
        self.conn.close()
        self.conn = None

    def get_table(self, name):
        if name not in self.metadata.tables:
            if name == '_synchronize':
                table = sa.Table(
                    name, self.metadata,
                    sa.Column('model', sa.Text, primary_key=True),
                    sa.Column('cid', sa.BIGINT),
                    sa.Column('updated', sa.DateTime),
                )
            else:
                table = sa.Table(
                    name, self.metadata,
                    sa.Column('key', sa.Text, primary_key=True),
                    sa.Column('hash', sa.Text, unique=True, index=True),
                    sa.Column('value', sa.LargeBinary),
                )
            table.create(checkfirst=True)
        return self.metadata.tables[name]

    def encode(self, name: str, value: object, primary_key=None) -> Optional[str]:
        # Make value msgpack serializable.
        hash_return = _hash_value(value)
        if hash_return is None:
            return None
        else:
            value, hashed = hash_return

        tmp_name = None
        if '.' in name:
            tmp_name = name.split('.')[0]

        table = self.get_table(name)

        if primary_key is not None:
            primary_key_table = self.get_table(tmp_name if tmp_name else name)
            query = sa.select([primary_key_table.c.key]).where(primary_key_table.c.key == primary_key)
        else:
            query = sa.select([table.c.key]).where(table.c.hash == hashed)
        key = self.conn.execute(query).scalar()

        if primary_key:
            stmt = insert(table).values(
                key=primary_key,
                hash=hashed,
                value=value
            )
            stmt = stmt.on_conflict_do_nothing()
            self.conn.execute(stmt)
            return primary_key
        if key is None:
            key = str(uuid.uuid4())
            self.conn.execute(table.insert(), {
                'key': key,
                'hash': hashed,
                'value': value,
            })
        return key

    def decode(self, name: str, key: str) -> object:
        table = self.get_table(name)
        query = sa.select([table.c.value]).where(table.c.key == key)
        value = self.conn.execute(query).scalar()
        value = msgpack.loads(value, raw=False)
        return value

    def get_sync_data(self, name: str) -> object:
        table = self.get_table('_synchronize')
        query = sa.select([table.c.cid]).where(table.c.model == name)
        value = self.conn.execute(query).scalar()
        return value

    def update_sync_data(self, name: str, cid: Any, time: datetime.datetime):
        table = self.get_table('_synchronize')
        query = insert(table).values(model=name, cid=cid, updated=time)
        query = query.on_conflict_do_update(
            index_elements=[table.c.model],
            set_=dict(cid=cid, updated=time)
        )
        self.conn.execute(query)

    def synchronize(self, name: str, value: Any, primary_key: str):
        table = self.get_table(name)
        hash_return = _hash_value(value)
        if hash_return is None:
            return None
        else:
            value, hashed = hash_return

        where_condition = sa.or_(table.c.key == primary_key, table.c.hash == hashed)
        select_query = sa.select([sa.func.count()]).select_from(table).where(where_condition)
        count = self.conn.execute(select_query).scalar()
        should_insert = True
        if count == 1:
            should_insert = False
            update_query = sa.update(table).values(key=primary_key, hash=hashed, value=value).where(where_condition)
            self.conn.execute(update_query)
        else:
            delete_query = sa.delete(table).where(where_condition)
            self.conn.execute(delete_query)

        if should_insert:
            query = insert(table).values(key=primary_key, hash=hashed, value=value)
            self.conn.execute(query)

    def first_time_sync(self) -> bool:
        table = self.get_table('_synchronize')
        count = self.conn.execute(sa.func.count(table.c.model)).scalar()
        return count == 0


def _hash_value(value):
    if isinstance(value, (list, tuple)):
        value = [_encode_value(k) for k in value if k is not None]
        if len(value) == 0:
            return None
    else:
        if value is None:
            return None
        value = _encode_value(value)
    value = msgpack.dumps(value, strict_types=True)
    return value, hashlib.sha1(value).hexdigest()


def _encode_value(value):
    if isinstance(value, (datetime.datetime, datetime.date)):
        return value.isoformat()
    if isinstance(value, decimal.Decimal):
        return float(value)
    if isinstance(value, tuple):
        return list(value)
    if isinstance(value, UUID):
        return str(value)
    return value


@commands.configure.register(Context, SqlAlchemyKeyMap)
def configure(context: Context, keymap: SqlAlchemyKeyMap):
    rc: RawConfig = context.get('rc')
    config: Config = context.get('config')
    dsn = rc.get('keymaps', keymap.name, 'dsn', required=True)
    ensure_data_dir(config.data_path)
    dsn = dsn.format(data_dir=config.data_path)
    if dsn.startswith('sqlite:///'):
        dsn = dsn.replace('sqlite:///', 'sqlite+spinta_sqlite:///')
    keymap.dsn = dsn


@commands.prepare.register(Context, SqlAlchemyKeyMap)
def prepare(context: Context, keymap: SqlAlchemyKeyMap):
    keymap.engine = sa.create_engine(keymap.dsn)
    keymap.metadata = sa.MetaData(keymap.engine)



from collections.abc import Iterable

from spinta.utils.schema import NA


def take(keys, *args):
    """Ultimate data extraction function

    Examples:

    +---------------------+----------------------------------------------------+
    | With `take`         | Without `take`                                     |
    +=====================+====================================================+
    | take(data)          | {k: v for k, v in data.items()                     |
    |                     |       if not k.startswith('_')} if data else NA    |
    +---------------------+----------------------------------------------------+
    | take('a.b', data)   | v = data                                           |
    |                     | for k in 'a.b'.split('.'):                         |
    |                     |     v = v[k]                                       |
    +---------------------+----------------------------------------------------+
    | take('a', d1, d2)   | d1 and d1.get('a', NA) or d2 and d2.get('a', NA)   |
    +---------------------+----------------------------------------------------+

    Examples:

        >>> d = {
        ...     '_a': 1,
        ...     '_b': 2,
        ...     'c_': 3,
        ...     'd_': 4,
        ... }

    Take specified reserved and all non-reserved keys:

        >>> take(['_a', all], d)
        {'_a': 1, 'c_': 3, 'd_': 4}

    """
    reserved = False
    if keys is all:
        keys = []
        reserved = True

    key = False
    if isinstance(keys, str):
        key = True
        keys = [keys]

    if not isinstance(keys, (list, tuple)):
        args = (keys,) + args
        keys = []

    args = [a for a in args if a]

    if all in keys:
        keys.remove(all)
        for arg in reversed(args):
            for k in arg:
                if not k.startswith('_') and k not in keys:
                    keys.append(k)

    if len(keys) == 0:
        return {
            k: v
            for a in reversed(args)
            for k, v in a.items()
            if v is not NA and (reserved or not k.startswith('_'))
        }
    else:
        keys = [k.split('.') for k in keys]

    data = {}
    for k in keys:
        for v in args:
            for x in k:
                if v and x in v:
                    v = v[x]
                else:
                    break
            else:
                if v is not NA:
                    if key:
                        return v
                    data['.'.join(k)] = v
                    break

    if key:
        return NA
    else:
        return data



from spinta.core.ufuncs import ufunc, Expr
from spinta.exceptions import PropertyNotFound, InvalidRequestQuery, SummaryWithMultipleProperties
from spinta.ufuncs.summaryenv.components import SummaryEnv, BBox


@ufunc.resolver(SummaryEnv, Expr, name='bbox')
def bbox(env, expr):
    args, kwargs = expr.resolve(env)
    if len(args) == 4:
        env.bbox = BBox(
            x_min=args[0],
            y_min=args[1],
            x_max=args[2],
            y_max=args[3]
        )
    else:
        raise InvalidRequestQuery(query="bbox", format="bbox(min_lon, min_lat, max_lon, max_lat)")


@ufunc.resolver(SummaryEnv, Expr, name='and')
def and_(env, expr):
    expr.resolve(env)


@ufunc.resolver(SummaryEnv, Expr, name='select')
def select(env, expr):
    args, kwargs = expr.resolve(env)
    prop = None
    for key in args:
        if prop is None and str(key) in env.model.properties:
            prop = env.model.properties[str(key)]
        elif prop is not None:
            raise SummaryWithMultipleProperties(env.model)
        else:
            raise PropertyNotFound(env.model, property=args[0])
    env.prop = prop




import sqlalchemy as sa
import sqlalchemy.exc

from spinta import commands
from spinta.components import Context
from spinta.datasets.backends.sql.components import Sql


@commands.wait.register(Context, Sql)
def wait(context: Context, backend: Sql, *, fail: bool = False) -> bool:
    rc = context.get('rc')
    dsn = rc.get('backends', backend.name, 'dsn', default=None)
    if dsn is None:
        return True
    engine = sa.create_engine(dsn)
    try:
        conn = engine.connect()
    except (sqlalchemy.exc.OperationalError, sqlalchemy.exc.DBAPIError):
        if fail:
            raise
        else:
            return False
    else:
        conn.close()
        engine.dispose()
        return True



from starlette.types import ASGIApp
from starlette.types import Receive
from starlette.types import Scope
from starlette.types import Send

from spinta.components import Context


class ContextMiddleware:
    """Adds `request.state.context`.

    There is a global `context`, where all heavy things are preloaded as
    startup. This preloading happens on Starlette's 'startup' event.

    `ContextMiddleware` creates a fork of global preloaded context for each
    request and assigns it to `request.state.context`. Forked context can be
    modified in each request without effecting global context.
    """

    def __init__(self, app: ASGIApp, *, context: Context) -> None:
        self.app = app
        self.context = context

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] in ["http", "websocket"]:
            with self.context.fork('request') as context:
                scope.setdefault('state', {})
                scope['state']['context'] = context
                await self.app(scope, receive, send)
        else:
            await self.app(scope, receive, send)



from spinta.accesslog import create_accesslog
from spinta.auth import AdminToken, get_client_id_from_name, get_clients_path
from spinta.auth import BearerTokenValidator
from spinta.auth import Token
from spinta.auth import create_client_access_token
from spinta.components import Context


def require_auth(context: Context, client: str = None):
    # TODO: probably commands should also use an existing token in order to
    #       track who changed what.
    if client is None:
        token = AdminToken()
    else:
        if client == 'default':
            config = context.get('config')
            client = get_client_id_from_name(get_clients_path(config), config.default_auth_client)
            client = client if client else config.default_auth_client
        token = create_client_access_token(context, client)
        token = Token(token, BearerTokenValidator(context))
    context.set('auth.token', token)
    context.attach('accesslog', create_accesslog, context, loaders=(
        context.get('store'),
        context.get("auth.token"),
    ))
